from __future__ import annotations

import inspect
import logging
from tenacity import retry, wait_exponential, stop_after_delay, before_sleep_log, retry_unless_exception_type, retry_if_exception_type, retry_if_exception
from typing import Optional, Union, Tuple, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from aiokeydb.v2.core import KeyDB, AsyncKeyDB
    from tenacity import WrappedFn


logger = logging.getLogger(__name__)

_excluded_funcs = ['parse_response']

class retry_if_type(retry_if_exception):
    """
    Retries if the exception is of the given type
    """
    def __init__(
        self, 
        exception_types: Union[
            Type[BaseException],
            Tuple[Type[BaseException], ...],
        ] = Exception,
        excluded_types: Union[
            Type[BaseException],
            Tuple[Type[BaseException], ...],
        ] = None,
    ):
        self.exception_types = exception_types
        self.excluded_types = excluded_types
        super().__init__(
            lambda e: isinstance(e, exception_types) and not isinstance(e, excluded_types)
        )

def get_retryable_wrapper(  
    max_attempts: int = 15,
    max_delay: int = 60,
    logging_level: int = logging.DEBUG,
    **kwargs,
) -> 'WrappedFn':
    
    """
    Creates a retryable decorator
    """
    from aiokeydb import exceptions as aiokeydb_exceptions
    from redis import exceptions as redis_exceptions
    return retry(
        wait = wait_exponential(multiplier = 0.5, min = 1, max = max_attempts),
        stop = stop_after_delay(max_delay),
        before_sleep = before_sleep_log(logger, logging_level),
        retry = retry_if_type(
            exception_types = (
                aiokeydb_exceptions.ConnectionError,
                aiokeydb_exceptions.TimeoutError,
                aiokeydb_exceptions.BusyLoadingError,
                redis_exceptions.ConnectionError,
                redis_exceptions.TimeoutError,
                redis_exceptions.BusyLoadingError,
            ),
            excluded_types = (
                aiokeydb_exceptions.AuthenticationError,
                aiokeydb_exceptions.AuthorizationError,
                aiokeydb_exceptions.InvalidResponse,
                aiokeydb_exceptions.ResponseError,
                aiokeydb_exceptions.NoScriptError,
                redis_exceptions.AuthenticationError,
                redis_exceptions.AuthorizationError,
                redis_exceptions.InvalidResponse,
                redis_exceptions.ResponseError,
                redis_exceptions.NoScriptError,
            )
        )
    )


def create_retryable_client(
    client: Type[Union[KeyDB, AsyncKeyDB]],

    max_attempts: int = 15,
    max_delay: int = 60,
    logging_level: int = logging.DEBUG,

    verbose: Optional[bool] = False,
    **kwargs
) -> Type[Union[KeyDB, AsyncKeyDB]]:
    """
    Creates a retryable client
    """
    if hasattr(client, '_is_retryable_wrapped'): return client
    decorator = get_retryable_wrapper(
        max_attempts = max_attempts,
        max_delay = max_delay,
        logging_level = logging_level,
        **kwargs
    )
    for attr in dir(client):
        if attr.startswith('_'): continue
        if attr in _excluded_funcs: continue
        attr_val = getattr(client, attr)
        if inspect.isfunction(attr_val) or inspect.iscoroutinefunction(attr_val):
            if verbose: logger.info(f'Wrapping {attr} with retryable decorator')
            setattr(client, attr, decorator(attr_val))
        
    setattr(client, '_is_retryable_wrapped', True)
    return client



    

# def wrap_retryable(
#     sess: 'KeyDBSession'
# ):
#     """
#     Wraps the keydb-session with a retryable decorator
#     """
#     if hasattr(sess, '_is_retryable_wrapped'): return sess
#     decorator = retry(
#         wait = wait_exponential(multiplier=0.5, min=1, max=15),
#         stop = stop_after_delay(60),
#         before_sleep = before_sleep_log(logger, logging.INFO)
#     )
#     for attr in dir(sess):
#         attr_val = getattr(sess, attr)
#         if inspect.isfunction(attr_val) or inspect.iscoroutinefunction(attr_val):
#             logger.info(f'Wrapping {attr} with retryable decorator')
#             setattr(sess, attr, decorator(attr_val))
        
#     setattr(sess, '_is_retryable_wrapped', True)
#     return sess



import os
import sys
import logging
import asyncio
import typing
import contextlib
import threading
import inspect
from itertools import chain
from queue import Empty, Full, Queue, LifoQueue
from urllib.parse import parse_qs, unquote, urlparse, ParseResult
from redis.connection import (
    URL_QUERY_ARGUMENT_PARSERS,
    Connection,
    UnixDomainSocketConnection,
    SSLConnection,
    Encoder,
    ConnectionPool as _ConnectionPool,
)

from redis.asyncio.connection import (
    Connection as AsyncConnection,
    UnixDomainSocketConnection as AsyncUnixDomainSocketConnection,
    SSLConnection as AsyncSSLConnection,
    Encoder as AsyncEncoder,
    ConnectionPool as _AsyncConnectionPool,
)

import aiokeydb.v2.exceptions as exceptions

# the functionality is available in 3.11.x but has a major issue before
# 3.11.3. See https://github.com/redis/redis-py/issues/2633
if sys.version_info >= (3, 11, 3):
    from asyncio import timeout as async_timeout
else:
    from async_timeout import timeout as async_timeout

from aiokeydb.v2.utils import set_ulimits

logger = logging.getLogger(__name__)

def parse_url(url: str, _is_async: bool = False):
    if not (
        url.startswith("keydb://")
        or url.startswith("keydbs://")
        or url.startswith("redis://")
        or url.startswith("rediss://")
        or url.startswith("unix://")
    ):
        raise ValueError(
            "KeyDB URL must specify one of the following "
            "schemes (keydb://, keydbs://, redis://, rediss://, unix://)"
        )

    url: ParseResult = urlparse(url)
    kwargs = {}

    for name, value in parse_qs(url.query).items():
        if value and len(value) > 0:
            value = unquote(value[0])
            parser = URL_QUERY_ARGUMENT_PARSERS.get(name)
            if parser:
                try:
                    kwargs[name] = parser(value)
                except (TypeError, ValueError) as e:
                    raise ValueError(f"Invalid value for `{name}` in connection URL.") from e
            else:
                kwargs[name] = value

    if url.username:
        kwargs["username"] = unquote(url.username)
    if url.password:
        kwargs["password"] = unquote(url.password)

    # We only support redis://, rediss:// and unix:// schemes.
    if url.scheme == "unix":
        if url.path:
            kwargs["path"] = unquote(url.path)
        kwargs["connection_class"] = AsyncUnixDomainSocketConnection if _is_async else UnixDomainSocketConnection

    else:  # implied:  url.scheme in ("redis", "rediss"):
        if url.hostname:
            kwargs["host"] = unquote(url.hostname)
        if url.port:
            kwargs["port"] = int(url.port)

        # If there's a path argument, use it as the db argument if a
        # querystring value wasn't specified
        if url.path and "db" not in kwargs:
            with contextlib.suppress(AttributeError, ValueError):
                kwargs["db"] = int(unquote(url.path).replace("/", ""))
        if url.scheme in {"rediss", "keydbs"}:
            kwargs["connection_class"] = AsyncSSLConnection if _is_async else SSLConnection

    return kwargs


_conn_valid_kwarg_keys: typing.Dict[str, typing.List[str]] = {}

def _extract_cls_init_kwargs(obj: object) -> typing.List[str]:
    """
    Extracts the kwargs that are valid for a connection class
    """
    argspec = inspect.getfullargspec(obj.__init__)
    _args = []
    for arg in argspec.args:
        _args.append(arg)
    for arg in argspec.kwonlyargs:
        _args.append(arg)
    
    # Check if subclass of Connection
    if hasattr(obj, "__bases__"):
        for base in obj.__bases__:
            _args.extend(_extract_cls_init_kwargs(base))

    return list(set(_args))

def filter_kwargs_for_connection(conn_cls: typing.Type[Connection], kwargs: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
    # sourcery skip: dict-comprehension
    """
    Filter out kwargs that aren't valid for a connection class
    """
    global _conn_valid_kwarg_keys
    if conn_cls.__name__ not in _conn_valid_kwarg_keys:
        _conn_valid_kwarg_keys[conn_cls.__name__] = _extract_cls_init_kwargs(conn_cls)
    return {k: v for k, v in kwargs.items() if k in _conn_valid_kwarg_keys[conn_cls.__name__]}


class ConnectionPool(_ConnectionPool):
    """
    Create a connection pool. ``If max_connections`` is set, then this
    object raises :py:class:`~redis.exceptions.ConnectionError` when the pool's
    limit is reached.

    By default, TCP connections are created unless ``connection_class``
    is specified. Use class:`.UnixDomainSocketConnection` for
    unix sockets.

    Any additional keyword arguments are passed to the constructor of
    ``connection_class``.
    """

    @classmethod
    def from_url(cls, url, **kwargs):
        """
        Return a connection pool configured from the given URL.

        For example::
            keydb://[[username]:[password]]@localhost:6379/0
            keydbs://[[username]:[password]]@localhost:6379/0
            redis://[[username]:[password]]@localhost:6379/0
            rediss://[[username]:[password]]@localhost:6379/0
            unix://[[username]:[password]]@/path/to/socket.sock?db=0

        Five URL schemes are supported:
        - `keydb://` creates a TCP socket connection.
        - `keydbs://` creates a SSL wrapped TCP socket connection.
        - `redis://` creates a TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/redis>
        - `rediss://` creates a SSL wrapped TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/rediss>
        - ``unix://``: creates a Unix Domain Socket connection.

        The username, password, hostname, path and all querystring values
        are passed through urllib.parse.unquote in order to replace any
        percent-encoded values with their corresponding characters.

        There are several ways to specify a database number. The first value
        found will be used:

            1. A ``db`` querystring option, e.g. redis://localhost?db=0
            2. If using the redis:// or rediss:// schemes, the path argument
               of the url, e.g. redis://localhost/0
            3. A ``db`` keyword argument to this function.

        If none of these options are specified, the default db=0 is used.

        All querystring options are cast to their appropriate Python types.
        Boolean arguments can be specified with string values "True"/"False"
        or "Yes"/"No". Values that cannot be properly cast cause a
        ``ValueError`` to be raised. Once parsed, the querystring arguments
        and keyword arguments are passed to the ``ConnectionPool``'s
        class initializer. In the case of conflicting arguments, querystring
        arguments always win.
        """
        url_options = parse_url(url, _is_async = False)
        if "connection_class" in kwargs:
            url_options["connection_class"] = kwargs["connection_class"]
        kwargs.update(url_options)
        return cls(**kwargs)
    

    def __init__(
        self, 
        connection_class: typing.Type[Connection] = Connection, 
        max_connections: typing.Optional[int] = None, 
        auto_pubsub: typing.Optional[bool] = True,
        pubsub_decode_responses: typing.Optional[bool] = True, # automatically set decode_responses to True
        auto_reset_enabled: typing.Optional[bool] = True,
        **connection_kwargs
    ):
        max_connections = max_connections or 2**31
        if not isinstance(max_connections, int) or max_connections < 0:
            raise ValueError('"max_connections" must be a positive integer')
        try:
            set_ulimits(max_connections)
        except Exception as e:
            logger.debug(f"Unable to set ulimits for connection: {e}")
        self.connection_class = connection_class
        # self.connection_kwargs = connection_kwargs
        self.connection_kwargs = filter_kwargs_for_connection(self.connection_class, connection_kwargs)
        self.max_connections = max_connections

        self._lock: threading.Lock = None
        self._created_connections: int = None
        self._available_connections: typing.List[typing.Type[Connection]] = None
        self._in_use_connections: typing.Set[typing.Type[Connection]] = None

        # a lock to protect the critical section in _checkpid().
        # this lock is acquired when the process id changes, such as
        # after a fork. during this time, multiple threads in the child
        # process could attempt to acquire this lock. the first thread
        # to acquire the lock will reset the data structures and lock
        # object of this pool. subsequent threads acquiring this lock
        # will notice the first thread already did the work and simply
        # release the lock.
        
        self._fork_lock: threading.Lock = threading.Lock()
        self.reset()
        self.encoder_class = self.connection_kwargs.get("encoder_class", Encoder)

        self.auto_pubsub = auto_pubsub
        self.pubsub_decode_responses = pubsub_decode_responses
        self._pubsub_encoder = self.get_pubsub_encoder() if self.auto_pubsub else None
        self._auto_reset_enabled = auto_reset_enabled

    @property
    def db_id(self):
        return self.connection_kwargs.get("db", 0)
    
    def with_db_id(self, db_id: int) -> 'ConnectionPool':
        """
        Return a new connection pool with the specified db_id.
        """
        kwargs = self.connection_kwargs.copy()
        kwargs["db"] = db_id
        return self.__class__(
            connection_class = self.connection_class,
            max_connections = self.max_connections,
            auto_pubsub = self.auto_pubsub,
            pubsub_decode_responses = self.pubsub_decode_responses,
            auto_reset_enabled = self._auto_reset_enabled,
            **kwargs
        )

    def reset(self):
        "Reset the pool"
        if not self._lock:
            self._lock = threading.Lock()
        
        self._created_connections = 0
        self._available_connections = []
        self._in_use_connections = set()

        # this must be the last operation in this method. while reset() is
        # called when holding _fork_lock, other threads in this process
        # can call _checkpid() which compares self.pid and os.getpid() without
        # holding any lock (for performance reasons). keeping this assignment
        # as the last operation ensures that those other threads will also
        # notice a pid difference and block waiting for the first thread to
        # release _fork_lock. when each of these threads eventually acquire
        # _fork_lock, they will notice that another thread already called
        # reset() and they will immediately release _fork_lock and continue on.
        self.pid = os.getpid()
    
    def get_connection(self, command_name, *keys, **options):
        "Get a connection from the pool"
        self._checkpid()
        with self._lock:
            try:
                connection = self._available_connections.pop()
            except IndexError:
                try:
                    connection = self.make_connection()
                except exceptions.ConnectionError as e:
                    if not self._auto_reset_enabled: raise e
                    logger.warning(f'Resetting Connection Pool: {self._created_connections}/{self.max_connections}')
                    self.reset_pool()
                    connection = self.make_connection()

            self._in_use_connections.add(connection)

            if command_name in {'PUBLISH', 'SUBSCRIBE', 'UNSUBSCRIBE'} and self.auto_pubsub:
                connection.encoder = self._pubsub_encoder

        try:
            # ensure this connection is connected to Redis
            connection.connect()
            # connections that the pool provides should be ready to send
            # a command. if not, the connection was either returned to the
            # pool before all data has been read or the socket has been
            # closed. either way, reconnect and verify everything is good.
            try:
                if connection.can_read():
                    raise ConnectionError("Connection has data")
            except (exceptions.ConnectionError, ConnectionError, OSError) as exc:
                connection.disconnect()
                connection.connect()
                if connection.can_read():
                    raise ConnectionError("Connection not ready") from exc
        except BaseException:
            # release the connection back to the pool so that we don't
            # leak it
            self.release(connection)
            raise

        return connection
    

    def get_pubsub_encoder(self):
        """
        Returns the pubsub encoder based on encoding settings
        """
        kwargs = self.connection_kwargs
        return self.encoder_class(
            encoding = kwargs.get("encoding", "utf-8"),
            encoding_errors = kwargs.get("encoding_errors", "strict"),
            decode_responses = self.pubsub_decode_responses or kwargs.get("decode_responses", True),
        )

    def disconnect(
        self, 
        inuse_connections: bool = True,
        raise_exceptions: bool = True,
        with_lock: bool = True,
    ):
        """
        Disconnects connections in the pool

        If ``inuse_connections`` is True, disconnect connections that are
        current in use, potentially by other threads. Otherwise only disconnect
        connections that are idle in the pool.
        """
        self._checkpid()
        if with_lock:
            self._lock.acquire()
        
        # with self._lock:
        if inuse_connections:
            connections = chain(
                self._available_connections, self._in_use_connections
            )
        else:
            connections = self._available_connections

        for connection in connections:
            try:
                connection.disconnect()
            except Exception as e:
                if raise_exceptions: raise e
        if with_lock:
            self._lock.release()
        
    def reset_pool(
        self, 
        inuse_connections: bool = True, 
        raise_exceptions: bool = False,
    ):
        """
        Resets the connection pool
        """
        self.disconnect(inuse_connections = inuse_connections, raise_exceptions = raise_exceptions, with_lock = False)
        self.reset()
    
    @property
    def _stats(self):
        return {
            "pid": self.pid,
            "created_connections": self._created_connections,
            "available_connections": len(self._available_connections),
            "in_use_connections": len(self._in_use_connections),
            "max_connections": self.max_connections,
        }

_ACP = typing.TypeVar("_ACP", bound="AsyncConnectionPool")

class AsyncConnectionPool(_AsyncConnectionPool):

    """
    Create a connection pool. ``If max_connections`` is set, then this
    object raises :py:class:`~redis.ConnectionError` when the pool's
    limit is reached.

    By default, TCP connections are created unless ``connection_class``
    is specified. Use :py:class:`~redis.UnixDomainSocketConnection` for
    unix sockets.

    Any additional keyword arguments are passed to the constructor of
    ``connection_class``.
    """

    @classmethod
    def from_url(cls: typing.Type[_ACP], url: str, **kwargs) -> _ACP:
        """
        Return a connection pool configured from the given URL.

        For example::
            keydb://[[username]:[password]]@localhost:6379/0
            keydbs://[[username]:[password]]@localhost:6379/0
            redis://[[username]:[password]]@localhost:6379/0
            rediss://[[username]:[password]]@localhost:6379/0
            unix://[[username]:[password]]@/path/to/socket.sock?db=0

        Five URL schemes are supported:
        - `keydb://` creates a TCP socket connection.
        - `keydbs://` creates a SSL wrapped TCP socket connection.
        - `redis://` creates a TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/redis>
        - `rediss://` creates a SSL wrapped TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/rediss>
        - ``unix://``: creates a Unix Domain Socket connection.

        The username, password, hostname, path and all querystring values
        are passed through urllib.parse.unquote in order to replace any
        percent-encoded values with their corresponding characters.

        There are several ways to specify a database number. The first value
        found will be used:
            1. A ``db`` querystring option, e.g. redis://localhost?db=0
            2. If using the redis:// or rediss:// schemes, the path argument
               of the url, e.g. redis://localhost/0
            3. A ``db`` keyword argument to this function.

        If none of these options are specified, the default db=0 is used.

        All querystring options are cast to their appropriate Python types.
        Boolean arguments can be specified with string values "True"/"False"
        or "Yes"/"No". Values that cannot be properly cast cause a
        ``ValueError`` to be raised. Once parsed, the querystring arguments
        and keyword arguments are passed to the ``ConnectionPool``'s
        class initializer. In the case of conflicting arguments, querystring
        arguments always win.
        """
        url_options = parse_url(url, _is_async = True)
        if "connection_class" in kwargs:
            url_options["connection_class"] = kwargs["connection_class"]
        kwargs.update(url_options)
        return cls(**kwargs)
    
    def __init__(
        self, 
        connection_class: typing.Type[AsyncConnection] = AsyncConnection, 
        max_connections: typing.Optional[int] = None, 
        auto_pubsub: typing.Optional[bool] = True,
        pubsub_decode_responses: typing.Optional[bool] = True, # automatically set decode_responses to True
        auto_reset_enabled: typing.Optional[bool] = True,
        **connection_kwargs
    ):
        max_connections = max_connections or 2**31
        if not isinstance(max_connections, int) or max_connections < 0:
            raise ValueError('"max_connections" must be a positive integer')
        try:
            set_ulimits(max_connections)
            # logger.info(f"Set ulimits to {max_connections}")
        except Exception as e:
            logger.debug(f"Unable to set ulimits for connection: {e}")
        self.connection_class: typing.Type[AsyncConnection] = connection_class
        self.connection_kwargs = filter_kwargs_for_connection(self.connection_class, connection_kwargs or {})
        # self.connection_kwargs: typing.Dict[str, typing.Any] = connection_kwargs or {}
        self.max_connections = max_connections

        self._lock: asyncio.Lock = None
        self._created_connections: int = None
        self._available_connections: typing.List[typing.Type[AsyncConnection]] = None
        self._in_use_connections: typing.Set[typing.Type[AsyncConnection]] = None

        # a lock to protect the critical section in _checkpid().
        # this lock is acquired when the process id changes, such as
        # after a fork. during this time, multiple threads in the child
        # process could attempt to acquire this lock. the first thread
        # to acquire the lock will reset the data structures and lock
        # object of this pool. subsequent threads acquiring this lock
        # will notice the first thread already did the work and simply
        # release the lock.
        
        self._fork_lock: threading.Lock = threading.Lock()
        self.reset()
        self.encoder_class = self.connection_kwargs.get("encoder_class", AsyncEncoder)

        self.auto_pubsub = auto_pubsub
        self.pubsub_decode_responses = pubsub_decode_responses
        self._pubsub_encoder = self.get_pubsub_encoder() if self.auto_pubsub else None
        self._auto_reset_enabled = auto_reset_enabled

    @property
    def db_id(self):
        return self.connection_kwargs.get("db", 0)
    
    def with_db_id(self, db_id: int) -> 'AsyncConnectionPool':
        """
        Return a new connection pool with the specified db_id.
        """
        kwargs = self.connection_kwargs.copy()
        kwargs["db"] = db_id
        return self.__class__(
            connection_class=self.connection_class,
            max_connections=self.max_connections,
            auto_pubsub=self.auto_pubsub,
            pubsub_decode_responses=self.pubsub_decode_responses,
            auto_reset_enabled=self._auto_reset_enabled,
            **kwargs,
        )
        

    def reset(self):
        if not self._lock:
            self._lock = asyncio.Lock()
        self._created_connections = 0
        self._available_connections = []
        self._in_use_connections = set()

        # this must be the last operation in this method. while reset() is
        # called when holding _fork_lock, other threads in this process
        # can call _checkpid() which compares self.pid and os.getpid() without
        # holding any lock (for performance reasons). keeping this assignment
        # as the last operation ensures that those other threads will also
        # notice a pid difference and block waiting for the first thread to
        # release _fork_lock. when each of these threads eventually acquire
        # _fork_lock, they will notice that another thread already called
        # reset() and they will immediately release _fork_lock and continue on.
        self.pid = os.getpid()
    
    def get_pubsub_encoder(self):
        """
        Returns the pubsub encoder based on encoding settings
        """
        kwargs = self.connection_kwargs
        return self.encoder_class(
            encoding = kwargs.get("encoding", "utf-8"),
            encoding_errors = kwargs.get("encoding_errors", "strict"),
            decode_responses = self.pubsub_decode_responses or kwargs.get("decode_responses", True),
        )
    

    async def get_connection(self, command_name, *keys, **options):
        """Get a connection from the pool"""
        self._checkpid()
        async with self._lock:
            try:
                connection: typing.Type[AsyncConnection] = self._available_connections.pop()
                # logger.info(f"{command_name} {keys} {options} | Got connection: {connection}: {connection.pid} {self._created_connections}/{self.max_connections}")
            except IndexError:
                try:
                    connection = self.make_connection()
                
                except exceptions.ConnectionError as e:
                    if not self._auto_reset_enabled: raise e

                    logger.warning(f'Resetting Connection Pool: {self._created_connections}/{self.max_connections}')
                    await self.reset_pool()
                    connection = self.make_connection()

            self._in_use_connections.add(connection)
            if command_name in {'PUBLISH', 'SUBSCRIBE', 'UNSUBSCRIBE'} and self.auto_pubsub:
                connection.encoder = self._pubsub_encoder

        try:
            # ensure this connection is connected to Redis
            await connection.connect()
            # connections that the pool provides should be ready to send
            # a command. if not, the connection was either returned to the
            # pool before all data has been read or the socket has been
            # closed. either way, reconnect and verify everything is good.
            try:
                if await connection.can_read_destructive():
                    raise ConnectionError("Connection has data") from None
            except (exceptions.ConnectionError, OSError, ConnectionError):
                await connection.disconnect()
                await connection.connect()
                if await connection.can_read_destructive():
                    raise ConnectionError("Connection not ready") from None
        except BaseException:
            # release the connection back to the pool so that we don't
            # leak it
            await self.release(connection)
            raise

        return connection
    
    async def disconnect(    
        self, 
        inuse_connections: bool = True,
        raise_exceptions: bool = False,
        with_lock: bool = True,
        return_connections: bool = False,
    ):
        """
        Disconnects connections in the pool

        If ``inuse_connections`` is True, disconnect connections that are
        current in use, potentially by other tasks. Otherwise only disconnect
        connections that are idle in the pool.
        """
        self._checkpid()
        if with_lock:
            await self._lock.acquire()
        if inuse_connections:
            connections: typing.Iterable[Connection] = chain(
                self._available_connections, self._in_use_connections
            )
        else:
            connections = self._available_connections
        resp = await asyncio.gather(
            *(connection.disconnect() for connection in connections),
            return_exceptions=True,
        )
        exc = next((r for r in resp if isinstance(r, BaseException)), None)
        if exc and raise_exceptions:
            raise exc
        if with_lock:
            self._lock.release()
        if return_connections:
            return [connection for connection, r in zip(connections, resp) if not isinstance(r, BaseException)]
        

    async def reset_pool(
        self, 
        inuse_connections: bool = True, 
        raise_exceptions: bool = False,
        with_lock: bool = False,
    ):
        """
        Resets the connection pool
        """
        await self.disconnect(inuse_connections = inuse_connections, raise_exceptions = raise_exceptions, with_lock = with_lock, return_connections = True)
        # logger.warning(f"Resetting Connection Pool: {self._created_connections}/{self.max_connections}")
        # for conn in conns:
        #     logger.error(f"Resetting Connection: {conn}: {conn.pid}")
        #     await self.release(conn)
        self.reset()


    @property
    def _stats(self):
        return {
            "pid": self.pid,
            "created_connections": self._created_connections,
            "available_connections": len(self._available_connections),
            "in_use_connections": len(self._in_use_connections),
            "max_connections": self.max_connections,
        }
    

## Blocking Pools

class BlockingConnectionPool(ConnectionPool):
    """
    Thread-safe blocking connection pool::

        >>> from redis.client import Redis
        >>> client = Redis(connection_pool=BlockingConnectionPool())

    It performs the same function as the default
    :py:class:`~redis.ConnectionPool` implementation, in that,
    it maintains a pool of reusable connections that can be shared by
    multiple redis clients (safely across threads if required).

    The difference is that, in the event that a client tries to get a
    connection from the pool when all of connections are in use, rather than
    raising a :py:class:`~redis.ConnectionError` (as the default
    :py:class:`~redis.ConnectionPool` implementation does), it
    makes the client wait ("blocks") for a specified number of seconds until
    a connection becomes available.

    Use ``max_connections`` to increase / decrease the pool size::

        >>> pool = BlockingConnectionPool(max_connections=10)

    Use ``timeout`` to tell it either how many seconds to wait for a connection
    to become available, or to block forever:

        >>> # Block forever.
        >>> pool = BlockingConnectionPool(timeout=None)

        >>> # Raise a ``ConnectionError`` after five seconds if a connection is
        >>> # not available.
        >>> pool = BlockingConnectionPool(timeout=5)
    """

    def __init__(
        self,
        max_connections: typing.Optional[int] = None,
        timeout: typing.Optional[int] = 20,
        connection_class: typing.Type[Connection] = Connection,
        queue_class: typing.Type[Queue] = LifoQueue,
        **connection_kwargs,
    ):
        max_connections = max_connections or 50
        self.queue_class = queue_class
        self.timeout = timeout
        self._connections: typing.List[Connection]
        connection_kwargs = filter_kwargs_for_connection(connection_class, connection_kwargs or {})
        super().__init__(
            connection_class = connection_class,
            max_connections = max_connections,
            **connection_kwargs,
        )

    @property
    def db_id(self):
        return self.connection_kwargs.get("db", 0)
    
    def with_db_id(self, db_id: int) -> 'BlockingConnectionPool':
        """
        Return a new connection pool with the specified db_id.
        """
        kwargs = self.connection_kwargs.copy()
        kwargs["db"] = db_id
        return self.__class__(
            max_connections=self.max_connections,
            timeout=self.timeout,
            connection_class=self.connection_class,
            queue_class=self.queue_class,
            auto_pubsub=self.auto_pubsub,
            pubsub_decode_responses=self.pubsub_decode_responses,
            auto_reset_enabled=self._auto_reset_enabled,
            **kwargs,
        )
    
    def reset(self):
        # Create and fill up a thread safe queue with ``None`` values.
        self.pool = self.queue_class(self.max_connections)
        while True:
            try:
                self.pool.put_nowait(None)
            except Full:
                break

        # Keep a list of actual connection instances so that we can
        # disconnect them later.
        self._connections: typing.List[typing.Type[Connection]] = []

        # this must be the last operation in this method. while reset() is
        # called when holding _fork_lock, other threads in this process
        # can call _checkpid() which compares self.pid and os.getpid() without
        # holding any lock (for performance reasons). keeping this assignment
        # as the last operation ensures that those other threads will also
        # notice a pid difference and block waiting for the first thread to
        # release _fork_lock. when each of these threads eventually acquire
        # _fork_lock, they will notice that another thread already called
        # reset() and they will immediately release _fork_lock and continue on.
        self.pid = os.getpid()

    def make_connection(self) -> typing.Type[Connection]:
        "Make a fresh connection."
        connection = self.connection_class(**self.connection_kwargs)
        self._connections.append(connection)
        return connection

    def get_connection(self, command_name, *keys, **options):
        """
        Get a connection, blocking for ``self.timeout`` until a connection
        is available from the pool.

        If the connection returned is ``None`` then creates a new connection.
        Because we use a last-in first-out queue, the existing connections
        (having been returned to the pool after the initial ``None`` values
        were added) will be returned before ``None`` values. This means we only
        create new connections when we need to, i.e.: the actual number of
        connections will only increase in response to demand.
        """
        # Make sure we haven't changed process.
        self._checkpid()

        # Try and get a connection from the pool. If one isn't available within
        # self.timeout then raise a ``ConnectionError``.
        connection = None
        try:
            connection = self.pool.get(block=True, timeout=self.timeout)
        except Empty as e:
            # Note that this is not caught by the redis client and will be
            # raised unless handled by application code. If you want never to
            if not self._auto_reset_enabled:
                raise ConnectionError(
                    "No connection available. Try enabling `auto_reset_enabled`."
                ) from e
            logger.warning(f'Resetting Pool: {len(self._connections)}/{self.pool.qsize()}/{self.max_connections} due to error: {e}')
            self.reset_pool()

        # If the ``connection`` is actually ``None`` then that's a cue to make
        # a new connection to add to the pool.
        if connection is None:
            connection = self.make_connection()

        try:
            # ensure this connection is connected to Redis
            connection.connect()
            # connections that the pool provides should be ready to send
            # a command. if not, the connection was either returned to the
            # pool before all data has been read or the socket has been
            # closed. either way, reconnect and verify everything is good.

            if command_name in {'PUBLISH', 'SUBSCRIBE', 'UNSUBSCRIBE'} and self.auto_pubsub:
                connection.encoder = self._pubsub_encoder

            try:
                if connection.can_read():
                    raise ConnectionError("Connection has data")
            except (exceptions.ConnectionError, OSError, ConnectionError) as exc:
                connection.disconnect()
                connection.connect()
                if connection.can_read():
                    raise ConnectionError("Connection not ready") from exc
        except BaseException:
            # release the connection back to the pool so that we don't leak it
            self.release(connection)
            raise

        return connection

    def release(self, connection):
        "Releases the connection back to the pool."
        # Make sure we haven't changed process.
        self._checkpid()
        if not self.owns_connection(connection):
            # pool doesn't own this connection. do not add it back
            # to the pool. instead add a None value which is a placeholder
            # that will cause the pool to recreate the connection if
            # its needed.
            connection.disconnect()
            self.pool.put_nowait(None)
            return

        # Put the connection back into the pool.
        with contextlib.suppress(Full):
            self.pool.put_nowait(connection)

    def disconnect(self):
        "Disconnects all connections in the pool."
        self._checkpid()
        for connection in self._connections:
            connection.disconnect()


    @property
    def _stats(self):
        return {
            "pid": self.pid,
            "available_connections": self.pool.qsize(),
            "in_use_connections": len(self._connections),
            "max_connections": self.max_connections,
        }

class AsyncBlockingConnectionPool(AsyncConnectionPool):
    """
    Thread-safe blocking connection pool::

        >>> from redis.client import Redis
        >>> client = Redis(connection_pool=BlockingConnectionPool())

    It performs the same function as the default
    :py:class:`~redis.ConnectionPool` implementation, in that,
    it maintains a pool of reusable connections that can be shared by
    multiple redis clients (safely across threads if required).

    The difference is that, in the event that a client tries to get a
    connection from the pool when all of connections are in use, rather than
    raising a :py:class:`~redis.ConnectionError` (as the default
    :py:class:`~redis.ConnectionPool` implementation does), it
    makes the client wait ("blocks") for a specified number of seconds until
    a connection becomes available.

    Use ``max_connections`` to increase / decrease the pool size::

        >>> pool = BlockingConnectionPool(max_connections=10)

    Use ``timeout`` to tell it either how many seconds to wait for a connection
    to become available, or to block forever:

        >>> # Block forever.
        >>> pool = BlockingConnectionPool(timeout=None)

        >>> # Raise a ``ConnectionError`` after five seconds if a connection is
        >>> # not available.
        >>> pool = BlockingConnectionPool(timeout=5)
    """

    def __init__(
        self,
        max_connections: typing.Optional[int] = None,
        timeout: typing.Optional[int] = 20,
        connection_class: typing.Type[AsyncConnection] = AsyncConnection,
        queue_class: typing.Type[asyncio.Queue] = asyncio.LifoQueue,
        **connection_kwargs,
    ):
        max_connections = max_connections or 50
        self.queue_class = queue_class
        self.timeout = timeout
        self._connections: typing.List[AsyncConnection]
        connection_kwargs = filter_kwargs_for_connection(connection_class, connection_kwargs or {})
        super().__init__(
            connection_class = connection_class,
            max_connections = max_connections,
            **connection_kwargs,
        )


    @property
    def db_id(self):
        return self.connection_kwargs.get("db", 0)
    
    def with_db_id(self, db_id: int) -> 'AsyncBlockingConnectionPool':
        """
        Return a new connection pool with the specified db_id.
        """
        kwargs = self.connection_kwargs.copy()
        kwargs["db"] = db_id
        return self.__class__(
            max_connections=self.max_connections,
            timeout=self.timeout,
            connection_class=self.connection_class,
            queue_class=self.queue_class,
            auto_pubsub=self.auto_pubsub,
            pubsub_decode_responses=self.pubsub_decode_responses,
            auto_reset_enabled=self._auto_reset_enabled,
            **kwargs,
        )
    
    def reset(self):
        # Create and fill up a thread safe queue with ``None`` values.
        self.pool = self.queue_class(self.max_connections)
        while True:
            try:
                self.pool.put_nowait(None)
            except asyncio.QueueFull:
                break

        # Keep a list of actual connection instances so that we can
        # disconnect them later.
        self._connections = []

        # this must be the last operation in this method. while reset() is
        # called when holding _fork_lock, other threads in this process
        # can call _checkpid() which compares self.pid and os.getpid() without
        # holding any lock (for performance reasons). keeping this assignment
        # as the last operation ensures that those other threads will also
        # notice a pid difference and block waiting for the first thread to
        # release _fork_lock. when each of these threads eventually acquire
        # _fork_lock, they will notice that another thread already called
        # reset() and they will immediately release _fork_lock and continue on.
        self.pid = os.getpid()

    def make_connection(self) -> typing.Type[AsyncConnection]:
        """Make a fresh connection."""
        connection = self.connection_class(**self.connection_kwargs)
        self._connections.append(connection)
        return connection

    async def get_connection(self, command_name, *keys, **options):
        """
        Get a connection, blocking for ``self.timeout`` until a connection
        is available from the pool.

        If the connection returned is ``None`` then creates a new connection.
        Because we use a last-in first-out queue, the existing connections
        (having been returned to the pool after the initial ``None`` values
        were added) will be returned before ``None`` values. This means we only
        create new connections when we need to, i.e.: the actual number of
        connections will only increase in response to demand.
        """
        # Make sure we haven't changed process.
        self._checkpid()

        # Try and get a connection from the pool. If one isn't available within
        # self.timeout then raise a ``ConnectionError``.
        connection = None
        try:
            async with async_timeout(self.timeout):
                connection = await self.pool.get()
        except (asyncio.QueueEmpty, asyncio.TimeoutError) as e:
            # Note that this is not caught by the redis client and will be
            # raised unless handled by application code. If you want never to
            # raise ConnectionError("No connection available.")
            if not self._auto_reset_enabled:
                raise ConnectionError(
                    "No connection available. Try enabling `auto_reset_enabled`"
                ) from e

            logger.warning(f'Resetting Pool: {len(self._connections)}/{self.pool.qsize()}/{self.max_connections} due to error: {e}')
            await self.reset_pool()

        # If the ``connection`` is actually ``None`` then that's a cue to make
        # a new connection to add to the pool.
        if connection is None:
            connection = self.make_connection()

        try:
            # ensure this connection is connected to Redis
            await connection.connect()
            # connections that the pool provides should be ready to send
            # a command. if not, the connection was either returned to the
            # pool before all data has been read or the socket has been
            # closed. either way, reconnect and verify everything is good.

            if command_name in {'PUBLISH', 'SUBSCRIBE', 'UNSUBSCRIBE'} and self.auto_pubsub:
                connection.encoder = self._pubsub_encoder

            try:
                if await connection.can_read_destructive():
                    raise ConnectionError("Connection has data") from None
            except (exceptions.ConnectionError, OSError, ConnectionError):
                await connection.disconnect()
                await connection.connect()
                if await connection.can_read_destructive():
                    raise ConnectionError("Connection not ready") from None
        
        except BaseException:
            # release the connection back to the pool so that we don't leak it
            await self.release(connection)
            raise

        return connection

    async def release(self, connection: AsyncConnection):
        """Releases the connection back to the pool."""
        # Make sure we haven't changed process.
        self._checkpid()
        if not self.owns_connection(connection):
            # pool doesn't own this connection. do not add it back
            # to the pool. instead add a None value which is a placeholder
            # that will cause the pool to recreate the connection if
            # its needed.
            await connection.disconnect()
            self.pool.put_nowait(None)
            return

        # Put the connection back into the pool.
        try:
            self.pool.put_nowait(connection)
        except asyncio.QueueFull:
            # perhaps the pool has been reset() after a fork? regardless,
            # we don't want this connection
            pass

    async def disconnect(self, inuse_connections: bool = True, raise_exceptions: bool = False):
        """Disconnects all connections in the pool."""
        self._checkpid()
        async with self._lock:
            resp = await asyncio.gather(
                *(connection.disconnect() for connection in self._connections),
                return_exceptions=True,
            )
            exc = next((r for r in resp if isinstance(r, BaseException)), None)
            if exc:
                raise exc

    @property
    def _stats(self):
        return {
            "pid": self.pid,
            "available_connections": self.pool.qsize(),
            "in_use_connections": len(self._connections),
            "max_connections": self.max_connections,
        }


import os
import gc
import json
import time
import asyncio
import typing
import anyio
import threading

from contextlib import asynccontextmanager, suppress
from lazyops.utils.logs import default_logger as logger
# from lazyops.utils import logger
from lazyops.utils.serialization import ObjectEncoder
from aiokeydb.v2.types.base import BaseModel, Field, validator
from aiokeydb.v2.types.base import KeyDBUri, lazyproperty
from aiokeydb.v2.types.session import KeyDBSession

from aiokeydb.v2.serializers import SerializerType
from aiokeydb.v2.client import KeyDBClient
from aiokeydb.v2.commands import AsyncScript

from aiokeydb.v2.connection import (
    ConnectionPool, 
    BlockingConnectionPool,
    AsyncConnectionPool,
    AsyncBlockingConnectionPool
)
from aiokeydb.v2.types.jobs import (
    Job,
    JobStatus,
    FunctionTracker,
    TERMINAL_STATUSES,
    UNSUCCESSFUL_TERMINAL_STATUSES,
)
import aiokeydb.v2.exceptions as exceptions
from aiokeydb.v2.exceptions import JobError
from aiokeydb.v2.utils.queue import (
    millis,
    now,
    seconds,
    uuid4,
    get_default_job_key,
    ensure_coroutine_function,
    get_hostname,
    concurrent_map,
    timer,
    build_batches,
)

from aiokeydb.v2.configs import settings
from aiokeydb.v2.utils import set_ulimits, get_ulimits
from aiokeydb.v2.backoff import default_backoff
from redis.asyncio.retry import Retry



class QueueStats(BaseModel):
    """
    Holds the state for the queue
    """
    
    name: str
    prefix: typing.Optional[str] = 'queue'
    uuid: typing.Optional[str] = Field(default_factory = uuid4)
    started: typing.Optional[int] = Field(default_factory = now)
    last_job: typing.Optional[int] = Field(default_factory = now)
    complete: typing.Optional[int] = 0
    failed: typing.Optional[int] = 0
    aborted: typing.Optional[int] = 0
    retried: typing.Optional[int] = 0
    num_workers: typing.Optional[int] = 0
    active_workers: typing.Optional[typing.List[str]] = []
    worker_attributes: typing.Optional[typing.Dict[str, typing.Dict[str, typing.Any]]] = {}

    @validator('uuid')
    def validate_uuid(cls, v):
        return uuid4() if v is None else v

    @lazyproperty
    def incomplete_key(self) -> str:
        return f"{self.prefix}:{self.name}:incomplete"
    
    @lazyproperty
    def queued_key(self) -> str:
        return f"{self.prefix}:{self.name}:queued"
    
    @lazyproperty
    def active_key(self) -> str:
        return f"{self.prefix}:{self.name}:active"
    
    @lazyproperty
    def scheduled_key(self) -> str:
        return f"{self.prefix}:{self.name}:scheduled"
    
    @lazyproperty
    def sweep_key(self) -> str:
        return f"{self.prefix}:{self.name}:sweep"
    
    @lazyproperty
    def stats_key(self) -> str:
        return f"{self.prefix}:{self.name}:stats"
    
    @lazyproperty
    def complete_key(self) -> str:
        return f"{self.prefix}:{self.name}:complete"
    
    @lazyproperty
    def queue_info_key(self) -> str:
        return f"{self.prefix}:{self.name}:queue_info"
    
    @lazyproperty
    def worker_token_key(self) -> str:
        return f"{self.prefix}:{self.name}:token"
    
    @lazyproperty
    def management_queue_key(self) -> str:
        return f"{self.prefix}:{self.name}:mtg"
    
    @lazyproperty
    def abort_id_prefix(self) -> str:
        return f"{self.prefix}:{self.name}:abort"
    
    @lazyproperty
    def heartbeat_key(self) -> str:
        return f"{self.prefix}:{self.name}:heartbeat"
    
    @lazyproperty
    def function_tracker_key(self) -> str:
        return f"{self.prefix}:{self.name}:functiontracker"
    
    @lazyproperty
    def queue_job_ids_key(self) -> str:
        return f"{self.prefix}:{self.name}:jobids"

    def create_namespace(self, namespace: str) -> str:
        return f"{self.prefix}:{self.name}:{namespace}"
    
    @lazyproperty
    def node_name(self) -> str:
        get_hostname()


class TaskQueue:
    """
    TaskQueue is used to interact with aiokeydb.

    queue_name: name of the job queue
    prefix: prefix for the job queue
    db: database id for the job queue

    serializer: serializer for the job queue. If it is not specified, the default
        serializer will be used.
        Expects the following methods:
            dump: lambda that takes a dictionary and outputs bytes (default json.dumps)
            load: lambda that takes bytes and outputs a python dictionary (default json.loads)
    
    max_concurrency: maximum concurrent operations.
        This throttles calls to `enqueue`, `job`, and `abort` to prevent the Queue
        from consuming too many KeyDB connections.

    client_mode: whether to run in client mode. If True, the queue will not
        create any keys in the database, and will not create any management
        keys. This is useful for running a worker in a client-only mode.
    
    uuid: uuid of the queue
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        queue_name: typing.Optional[str] = None,
        prefix: typing.Optional[str] = None,
        db: typing.Optional[int] = None,
        serializer: typing.Optional[typing.Union[object, SerializerType]] = None,
        max_concurrency: typing.Optional[int] = None,
        max_broadcast_concurrency: typing.Optional[int] = None,
        client_mode: typing.Optional[bool] = False,
        debug_enabled: typing.Optional[bool] = None,
        uuid: typing.Optional[str] = None,
        metrics_func: typing.Optional[typing.Callable] = None,
        verbose_results: typing.Optional[bool] = False,
        truncate_logs: typing.Optional[bool] = True,
        logging_max_length: typing.Optional[int] = 500,
        silenced_functions: typing.Optional[typing.List[str]] = None, # Don't log these functions
        heartbeat_ttl: typing.Optional[int] = None, # 10,
        is_leader_process: typing.Optional[bool] = None,
        function_tracker_enabled: typing.Optional[bool] = None,
        function_tracker_ttl: typing.Optional[int] = None,
        **kwargs
    ): # sourcery skip: low-code-quality
        self.settings = settings
        self.prefix = prefix if prefix is not None else self.settings.worker.prefix
        self.queue_name = queue_name if queue_name is not None else self.settings.worker.queue_name
        self.serializer = serializer if serializer is not None else self.settings.worker.job_serializer.get_serializer()
        
        self._ctx_kwargs = {
            'db_id': db if db is not None else self.settings.worker.db,
            **kwargs        
        }

        self.client_mode = client_mode
        self.max_concurrency = max_concurrency or self.settings.worker.max_concurrency
        self.max_broadcast_concurrency = max_broadcast_concurrency or self.settings.worker.max_broadcast_concurrency or 3
        
        self.debug_enabled = debug_enabled if debug_enabled is not None else self.settings.worker.debug_enabled
        self.function_tracker_enabled = function_tracker_enabled if function_tracker_enabled is not None else self.settings.worker.function_tracker_enabled
        self.function_tracker_ttl = function_tracker_ttl if function_tracker_ttl is not None else self.settings.worker.function_tracker_ttl

        self._stats = QueueStats(
            name = self.queue_name,
            prefix = self.prefix,
            uuid = uuid,
        )
        self.uuid = self._stats.uuid
        self._version = None
        self._schedule_script: AsyncScript = None
        self._enqueue_script: AsyncScript = None
        self._cleanup_script: AsyncScript = None

        self._before_enqueues = {}
        self._op_sem = asyncio.Semaphore(self.max_concurrency)
        self._metrics_func = metrics_func

        self.verbose_results = verbose_results
        self.truncate_logs = truncate_logs
        self.logging_max_length = logging_max_length
        self.heartbeat_ttl = heartbeat_ttl or self.settings.worker.heartbeat_interval
        self._worker_name: str = None
        self.is_leader_process = is_leader_process if is_leader_process is not None else (self.settings.worker.is_leader_process if self.settings.worker.is_leader_process is not None else True)
        self.queue_pid: int = os.getpid()
        if silenced_functions:
            self.add_silenced_functions(*silenced_functions)
    
    @lazyproperty
    def _lock(self) -> asyncio.Lock:
        """
        Returns a lock for the queue.
        """
        return asyncio.Lock()
    
    @lazyproperty
    def ctx(self) -> KeyDBSession:
        """
        Returns the KeyDBSession for the queue.
        """
        if 'socket_connect_timeout' not in self._ctx_kwargs:
            self._ctx_kwargs['socket_connect_timeout'] = self.settings.worker.socket_connect_timeout
        if 'socket_keepalive' not in self._ctx_kwargs:
            self._ctx_kwargs['socket_keepalive'] = self.settings.worker.socket_keepalive
        if 'health_check_interval' not in self._ctx_kwargs:
            self._ctx_kwargs['health_check_interval'] = self.heartbeat_ttl
        
        return KeyDBClient.create_session(
            name = self.queue_name,
            serializer = False,
            cache_enabled = False,
            decode_responses = False,
            auto_pubsub = False,
            set_current = False,

            max_connections = self.max_concurrency * 10,
            pool_class = ConnectionPool,

            amax_connections = self.max_concurrency ** 2,
            apool_class = AsyncConnectionPool,
            **self._ctx_kwargs
        )

    async def _get_stats(self, log_stats: bool = False, include_jobs: bool = True, include_conn_kwargs: bool = True, include_retries: bool = False):
        """
        Fetches the stats for the queue.
        """
        _stats = await self.ctx._async_get_stats()
        _stats['version'] = self.version
        _stats['workers'] = {
            'num': self.num_workers,
            'active': self.active_workers,
        }
        _stats['queue'] = await self.info()
        if not include_jobs:
            _stats['queue']['jobs'] = len(_stats['queue'].get('jobs', []))
        if include_conn_kwargs: 
            _stats['connection_kwargs'] = {**self.ctx.async_client.connection_pool.connection_kwargs, **self._ctx_kwargs}
            if not include_retries:
                _stats['connection_kwargs']['retry_on_error'] = len(_stats['connection_kwargs']['retry_on_error'])
            
            # Remove password
            _ = _stats['connection_kwargs'].pop('password', None)

        if log_stats: 
            self.logger(kind = 'stats').info(f"{_stats}")
            if _stats.get('max_connections_used') > 95.0:
                self.logger(kind = 'stats').warning(f"Nearing Maximum connection client Limit: {_stats.get('max_connections_used')}")
        return _stats
    
    @lazyproperty
    def version(self):
        """
        Returns the version of the redis server.
        """
        info = self.ctx.info()
        return tuple(int(i) for i in info["redis_version"].split("."))
    
    @lazyproperty
    def uri(self) -> KeyDBUri:
        """
        Returns the connection for the queue.
        """
        return self.ctx.uri

    @lazyproperty
    def db_id(self):
        """
        Returns the database id for the queue.
        """
        return self.ctx.db_id
    
    @property
    def _should_debug_log(self):
        """
        Returns whether the queue should debug log.
        """
        return self.debug_enabled and self.is_leader_process
    

    def is_silenced_function(
        self,
        name: str,
        stage: typing.Optional[str] = None,
    ) -> bool:
        """
        Checks if a function is silenced
        """
        return self.settings.worker.is_silenced_function(name, stage = stage)


    def add_silenced_functions(self, *functions):
        """
        Adds functions to the list of silenced functions.
        """
        for func in functions:
            self.settings.worker.add_function_to_silenced(func)

    def logger(self, job: 'Job' = None, kind: str = "enqueue", job_id: typing.Optional[str] = None,):
        """
        Returns a logger for the queue.
        """
        if job:
            return logger.bind(
                job_id = job.id,
                status = job.status,
                worker_name = self._worker_name,
                queue_name = getattr(job.queue, 'queue_name', None) or 'unknown queue',
                kind = kind,
            )
        elif job_id:
            return logger.bind(
                job_id = job_id,
                worker_name = self._worker_name,
                queue_name = self.queue_name,
                kind = kind,
            )
        else:
            return logger.bind(
                worker_name = self._worker_name,
                queue_name = self.queue_name,
                kind = kind,
            )
    
    @property
    def job_id_prefix(self) -> str:
        """
        Returns the job id prefix.
        """
        return f"queue:{self.queue_name}:{self.settings.worker.job_prefix}"

    def job_id(self, job_key):
        """
        Returns the job id for a job key.
        """
        return f"{self.job_id_prefix}:{job_key}" if \
            self.job_id_prefix not in job_key else job_key
    
    def register_before_enqueue(self, callback):
        """
        Registers a callback to run before enqueue.
        """
        self._before_enqueues[id(callback)] = callback

    def unregister_before_enqueue(self, callback):
        """
        Unregisters a callback to run before enqueue.
        """
        self._before_enqueues.pop(id(callback), None)

    async def _before_enqueue(self, job: Job):
        """
        Helper function to run before enqueue.
        """
        for cb in self._before_enqueues.values():
            await cb(job)

    @asynccontextmanager
    async def pipeline(self, transaction: bool = True, shard_hint: typing.Optional[str] = None, raise_on_error: typing.Optional[bool] = None):
        """
        Context manager to run a pipeline.

        This context manager is a wrapper around the redis pipeline context manager to ensure that it is
        properly deleted.
        """
        yield self.ctx.async_pipeline
        # pipe = self.ctx.async_client.pipeline(transaction = transaction, shard_hint = shard_hint)
        # pipe = self.ctx.async_pipeline
        # yield pipe
        # await pipe.execute(raise_on_error = raise_on_error)
        # await pipe.close(close_connection_pool = False)
        # del pipe

    def serialize(self, job: Job):
        """
        Dumps a job.
        """
        return self.serializer.dumps(job.to_dict())

    def deserialize(self, job_bytes: bytes):
        """
        Deserializes a job.
        """
        if not job_bytes: return None
        job_dict: typing.Dict = self.serializer.loads(job_bytes)
        assert (
                job_dict.pop("queue") == self.queue_name
        ), f"Job {job_dict} fetched by wrong queue: {self.queue_name}"
        return Job(**job_dict, queue=self)

    async def disconnect(self):
        """
        Disconnects the queue.
        """
        await self.ctx.aclose()

    
    async def prepare_server(self):
        """
        Prepares the keydb server to ensure that the maximum number of concurrent
        connections are available.
        """
        self.ctx.config['transaction'] = True
        info = await self.ctx.async_info()
        await self.prepare_for_broadcast()
        if 'maxclients' not in info:
            if self._should_debug_log: logger.warning(f'Unable to configure the maxclients to {self.max_concurrency}: maxclients not supported: {info}')
            return

        # Each Worker spawns a connection
        min_connections = ((self.num_workers or 1) * (self.max_concurrency + self.max_broadcast_concurrency)) * 10

        curr_max_connections = info['maxclients']
        curr_connected = info['connected_clients']

        new_max_connections = max(
            min_connections,
            (curr_connected + min_connections),
            curr_max_connections
        )
        curr_ulimits = get_ulimits()
        if curr_ulimits < new_max_connections and self._should_debug_log:
            logger.debug(f'The maximum number of concurrent connections may not be supported as ulimits: {curr_ulimits} < desired max connections: {new_max_connections}')

        if self._should_debug_log: logger.info(f'Configuring Server: Current Max Connections: {curr_max_connections}, Current Connected: {curr_connected}, Min Connections: {min_connections} -> New Max: {new_max_connections}')
        if new_max_connections > curr_max_connections:
            try:
                await self.ctx.config_set('maxclients', new_max_connections)
                info = await self.ctx.async_info()
                if 'maxclients' in info:
                    new_set_max_connections = info['maxclients']
                    if self._should_debug_log: logger.debug(f'New Max Connections: {new_set_max_connections}')
            except Exception as e:
                if self._should_debug_log:
                    logger.warning(f'Unable to configure the maxclients to {new_max_connections}: {e}')
        # logger.info(f'Pool Class: {self.ctx.client_pools.apool.__class__.__name__}')


    """
    Primary APIs
    """

    async def schedule(self, lock: int = 1, worker_id: typing.Optional[str] = None):
        """
        Schedule jobs.
        """
        if not self._schedule_script:
            self._schedule_script = self.ctx.async_client.register_script(
                """
                if redis.call('EXISTS', KEYS[1]) == 0 then
                    redis.call('SETEX', KEYS[1], ARGV[1], 1)
                    local jobs = redis.call('ZRANGEBYSCORE', KEYS[2], 1, ARGV[2])

                    if next(jobs) then
                        local scores = {}
                        for _, v in ipairs(jobs) do
                            table.insert(scores, 0)
                            table.insert(scores, v)
                        end
                        redis.call('ZADD', KEYS[2], unpack(scores))
                        redis.call('RPUSH', KEYS[3], unpack(jobs))
                    end

                    return jobs
                end
                """
            )

        scheduled_key = self.scheduled_key if worker_id is None else f"{self.scheduled_key}:{worker_id}"
        incomplete_key = self.incomplete_key if worker_id is None else f"{self.incomplete_key}:{worker_id}"
        queued_key = self.queued_key if worker_id is None else f"{self.queued_key}:{worker_id}"

        return await self._schedule_script(
            keys=[scheduled_key, incomplete_key, queued_key],
            args=[lock, seconds(now())],
        )

    async def sweep(self, lock: int = 60, worker_id: typing.Optional[str] = None):
        """
        Sweep jobs.
        """
        # sourcery skip: low-code-quality
        if not self._cleanup_script:
            self._cleanup_script = self.ctx.async_client.register_script(
                """
                if redis.call('EXISTS', KEYS[1]) == 0 then
                    redis.call('SETEX', KEYS[1], ARGV[1], 1)
                    return redis.call('LRANGE', KEYS[2], 0, -1)
                end
                """
            )

        active_key = self.active_key if worker_id is None else f"{self.active_key}:{worker_id}"
        sweep_key = self.sweep_key if worker_id is None else f"{self.sweep_key}:{worker_id}"
        incomplete_key = self.incomplete_key if worker_id is None else f"{self.incomplete_key}:{worker_id}"

        job_ids = await self._cleanup_script(
            # keys = [self.sweep_key, self.active_key] if worker_id is None else [self.sweep_key, f"{self.active_key}:{worker_id}"],
            keys = [sweep_key, active_key],
            args = [lock], 
            client = self.ctx.async_client,
        )

        swept = []
        if job_ids:
            for job_id, job_bytes in zip(job_ids, await self.ctx.async_mget(job_ids)):
                try:
                    job = self.deserialize(job_bytes)
                except Exception as e:
                    job = None
                    self.logger(kind = "sweep").warning(f"Unable to deserialize job {job_id}: {e}")
                if not job: 
                    swept.append(job_id)
                    # async with self.pipeline(transaction = True) as pipe:
                    async with self.ctx.async_client.pipeline(transaction = True) as pipe:
                        await (
                            pipe.lrem(active_key, 0, job_id)
                            .zrem(incomplete_key, job_id)
                            .execute()
                        )
                        # await (
                        #     pipe.lrem(self.active_key, 0, job_id)
                        #     .zrem(self.incomplete_key, job_id)
                        #     .execute()
                        # )
                    self.logger(kind = "sweep").info(f"Sweeping missing job {job_id}")
                elif job.status != JobStatus.ACTIVE or job.stuck:
                    swept.append(job_id)
                    await job.finish(JobStatus.ABORTED, error="swept")
                    if self.is_silenced_function(job.function, stage = 'sweep'):
                        pass
                    elif self.verbose_results:
                        self.logger(job=job, kind = "sweep").info(f"☇ duration={job.duration('total')}ms, node={self.node_name}, func={job.function}, result={job.result}")
                    elif self.truncate_logs:
                        job_result = (
                            f'{str(job.result)[:self.logging_max_length]}...'
                            if self.logging_max_length
                            else str(job.result)
                        )
                        self.logger(job=job, kind = "sweep").info(f"☇ duration={job.duration('total')}ms, node={self.node_name}, func={job.function}, result={job_result}")
                    else:
                        self.logger(job=job, kind = "sweep").info(f"☇ duration={job.duration('total')}ms, node={self.node_name}, func={job.function}")
        gc.collect()
        return swept
    
    async def sweep_job(self, job_id: typing.Optional[str] = None, worker_id: typing.Optional[str] = None, job: typing.Optional[Job] = None, verbose: typing.Optional[bool] = None):
        # sourcery skip: low-code-quality
        """
        Sweep a single job.
        """
        assert job_id or job, "Must provide either job_id or job"
        job = await self._get_job_by_id(job_id) if job is None else job
        job_id = job_id if job_id is not None else job.id
        verbose = verbose if verbose is not None else self.verbose_results
        if not job:
            if verbose: self.logger(kind = "sweep").info(f"Sweeping missing job {job_id}")
        elif job.status != JobStatus.ACTIVE or job.stuck:
            await job.finish(JobStatus.ABORTED, error="swept")
            if self.is_silenced_function(job.function, stage = 'sweep'):
                pass
            elif verbose is False:
                self.logger(job=job, kind = "sweep").info(f"☇ duration={job.duration('total')}ms, node={self.node_name}, func={job.function}")
            elif self.verbose_results:
                self.logger(job=job, kind = "sweep").info(f"☇ duration={job.duration('total')}ms, node={self.node_name}, func={job.function}, result={job.result}")
            elif self.truncate_logs:
                job_result = (
                    f'{str(job.result)[:self.logging_max_length]}...'
                    if self.logging_max_length
                    else str(job.result)
                )
                self.logger(job=job, kind = "sweep").info(f"☇ duration={job.duration('total')}ms, node={self.node_name}, func={job.function}, result={job_result}")
                
        active_key = self.active_key if worker_id is None else f"{self.active_key}:{worker_id}"
        incomplete_key = self.incomplete_key if worker_id is None else f"{self.incomplete_key}:{worker_id}"
        async with self.ctx.async_client.pipeline(transaction = True) as pipe:
            await (
                pipe.lrem(active_key, 0, job_id)
                .zrem(incomplete_key, job_id)
                .execute()
            )
        return True

    async def listen(self, job_keys: typing.List[str], callback: typing.Callable, timeout: int = 10):
        """
        Listen to updates on jobs.

        job_keys: sequence of job keys
        callback: callback function, if it returns truthy, break
        timeout: if timeout is truthy, wait for timeout seconds
        """
        pubsub = self.ctx.async_client.pubsub(retryable=True)
        
        job_ids = [self.job_id(job_key) for job_key in job_keys]
        await pubsub.subscribe(*job_ids)

        async def listen():
            async for message in pubsub.listen():
                if message["type"] == "message":
                    job_id = message["channel"].decode("utf-8")
                    job_key = Job.key_from_id(job_id)
                    status = JobStatus[message["data"].decode("utf-8").upper()]
                    if asyncio.iscoroutinefunction(callback):
                        stop = await callback(job_key, status)
                    else:
                        stop = callback(job_key, status)
                    if stop: break

        try:
            if timeout:
                await asyncio.wait_for(listen(), timeout)
            else:
                await listen()
        finally:
            await pubsub.unsubscribe(*job_ids)

    async def notify(self, job: Job):
        """
        Notify subscribers of a job update.
        """
        await self.ctx.async_client.publish(job.id, job.status)
        await job.run_job_callback()


    async def update(self, job: Job):
        """
        Update a job.
        """
        job.touched = now()
        await self.ctx.async_set(job.id, self.serialize(job))
        await self.notify(job)
        if self.function_tracker_enabled: 
            await self.track_job_id(job)

    async def job(self, job_key: str) -> Job:
        """
        Fetch a Job by key.
        """
        job_id = self.job_id(job_key)
        return await self._get_job_by_id(job_id)
    
    async def job_exists(self, job_key: str) -> bool:
        """
        Check if a job exists
        """
        job_id = self.job_id(job_key)
        return await self.ctx.async_exists(job_id)

    async def _get_job_by_id(self, job_id):
        async with self._op_sem:
            return self.deserialize(await self.ctx.async_get(job_id))

    async def wait_for_job_completion(self, job_id: str, results_only: typing.Optional[bool] = False, interval: float = 0.5) -> typing.Any:
        """
        Waits for a job completion
        """
        job = await self.job(job_id)
        if not job: raise ValueError(f"Job {job_id} not found")
        while job.status not in TERMINAL_STATUSES:
            await asyncio.sleep(interval)
            job = await self.job(job_id)
        if not results_only: return job
        return job.result if job.status == JobStatus.COMPLETE else job.error


    async def abort(self, job: Job, error: typing.Any, ttl: int = 5):
        """
        Abort a job.
        """
        async with self._op_sem:
            async with self.ctx.async_client.pipeline(transaction = True, retryable = True) as pipe:
            # async with self.pipeline(transaction = True) as pipe:
                dequeued, *_ = await (
                    pipe.lrem(job.queued_key, 0, job.id)
                    .zrem(job.incomplete_key, job.id)
                    .expire(job.id, ttl + 1)
                    .setex(job.abort_id, ttl, error)
                    .execute()
                )

            if dequeued:
                await job.finish(JobStatus.ABORTED, error = error)
                await self.ctx.async_delete(job.abort_id)
            else:
                await self.ctx.async_lrem(job.active_key, 0, job.id)
            
            await self.track_job(job)


    async def retry(self, job: Job, error: typing.Any):
        """
        Retry an errored job.
        """
        job_id = job.id
        job.status = JobStatus.QUEUED
        job.error = error
        job.completed = 0
        job.started = 0
        job.progress = 0
        job.touched = now()
        next_retry_delay = job.next_retry_delay()

        # async with self.pipeline(transaction = True) as pipe:
        async with self.ctx.async_client.pipeline(transaction=True, retryable = True) as pipe:
            pipe = pipe.lrem(job.active_key, 1, job_id)
            pipe = pipe.lrem(job.queued_key, 1, job_id)
            if next_retry_delay:
                scheduled = time.time() + next_retry_delay
                pipe = pipe.zadd(job.incomplete_key, {job_id: scheduled})
            else:
                pipe = pipe.zadd(job.incomplete_key, {job_id: job.scheduled})
                pipe = pipe.rpush(job.queued_key, job_id)
            await pipe.set(job_id, self.serialize(job)).execute()
            self.retried += 1
            await self.notify(job)
            if not self.is_silenced_function(job.function, stage = 'retry'):
                self.logger(job=job, kind = "retry").info(f"↻ duration={job.duration('running')}ms, node={self.node_name}, func={job.function}, error={job.error}")

    async def finish(
        self, 
        job: Job, 
        status: typing.Union[JobStatus, str], 
        *, 
        result: typing.Any = None, 
        error: typing.Any = None
    ):
        """
        Publish a job result.
        """
        job_id = job.id
        job.status = status
        job.result = result
        job.error = error
        job.completed = now()

        if status == JobStatus.COMPLETE:
            job.progress = 1.0

        async with self.ctx.async_client.pipeline(transaction=True, retryable = True) as pipe:
        # async with self.pipeline(transaction = True) as pipe:
            pipe = pipe.lrem(job.active_key, 1, job_id).zrem(job.incomplete_key, job_id)
            if job.ttl > 0:
                pipe = pipe.setex(job_id, job.ttl, self.serialize(job))
            elif job.ttl == 0:
                pipe = pipe.set(job_id, self.serialize(job))
            else:
                pipe.delete(job_id)

            await pipe.execute()

            if status == JobStatus.COMPLETE:
                self.complete += 1
            elif status == JobStatus.FAILED:
                self.failed += 1
            elif status == JobStatus.ABORTED:
                self.aborted += 1

            await self.notify(job)
            if self.debug_enabled:
                self.logger(job=job, kind = "finish").info(f"Finished {job}")
            elif self.is_silenced_function(job.function, stage = 'finish'):
                pass
            elif self.verbose_results or not self.truncate_logs:
                self.logger(job=job, kind="finish").info(f"● duration={job.duration('total')}ms, node={self.node_name}, func={job.function}")
            else:
                job_result = (
                    f'{str(job.result)[:self.logging_max_length]}...'
                    if self.logging_max_length
                    else str(job.result)
                )
                self.logger(job=job, kind="finish").info(f"● duration={job.duration('total')}ms, node={self.node_name}, func={job.function}, result={job_result}")
            
            await self.track_job(job)

    async def _dequeue_job_id(
        self,
        timeout: int = 0,
        postfix: str = None,
    ):
        """
        Dequeue a job from the queue.
        """
        queued_key, active_key = self.queued_key, self.active_key
        if postfix:
            queued_key = f'{self.queued_key}:{postfix}'
            active_key = f'{self.active_key}:{postfix}'

        if self.version < (6, 2, 0):
            return await self.ctx.async_brpoplpush(
                queued_key, 
                active_key, 
                timeout
            )
        return await self.ctx.async_blmove(
            source_list = queued_key, 
            destination_list = active_key, 
            source = "RIGHT", 
            destination = "LEFT", 
            timeout = timeout
        )

    async def dequeue(self, timeout = 0, worker_id: str = None, worker_name: str = None):
        """
        Dequeue a job from the queue.
        """
        job_id = await self._dequeue_job_id(timeout) if \
            worker_id is None and worker_name is None else None
        
        if job_id is None and worker_id:
            job_id = await self._dequeue_job_id(timeout, worker_id)
            
        if job_id is None and worker_name:
            job_id = await self._dequeue_job_id(timeout, worker_name)
        
        if job_id is not None:
            # logger.info(f'Fetched job id {job_id}: {queued_key}: {active_key}')
            return await self._get_job_by_id(job_id)

        if not worker_id and not worker_name: self.logger(kind="dequeue").info("Dequeue timed out")
        return None
    
    
    async def _deferred(
        self, 
        job_or_func: typing.Union[Job, str],
        wait_time: float = 10.0, 
        **kwargs
    ):
        """
        Defer a job by instance or string after a certain amount of time
        """
        await asyncio.sleep(wait_time)
        return await self.enqueue(job_or_func, **kwargs)

    async def defer(
        self, 
        job_or_func: typing.Union[Job, str],
        wait_time: float = 10.0, 
        **kwargs
    ):
        """
        Enqueue a job by instance or string after a certain amount of time

        Kwargs can be arguments of the function or properties of the job.
        If a job instance is passed in, it's properties are overriden.

        If the job has already been enqueued, this returns None.
        will execute asyncronously and return immediately.
        """
        asyncio.create_task(self._deferred(job_or_func, wait_time = wait_time, **kwargs))


    async def enqueue(
        self, 
        job_or_func: typing.Union[Job, str, typing.Callable],
        **kwargs
    ):
        """
        Enqueue a job by instance or string.

        Kwargs can be arguments of the function or properties of the job.
        If a job instance is passed in, it's properties are overriden.

        If the job has already been enqueued, this returns None.
        """
        job = Job.from_kwargs(job_or_func = job_or_func, **kwargs)

        if not self._enqueue_script:
            self._enqueue_script = self.ctx.async_client.register_script(
                """
                if not redis.call('ZSCORE', KEYS[1], KEYS[2]) and redis.call('EXISTS', KEYS[4]) == 0 then
                    redis.call('SET', KEYS[2], ARGV[1])
                    redis.call('ZADD', KEYS[1], ARGV[2], KEYS[2])
                    if ARGV[2] == '0' then redis.call('RPUSH', KEYS[3], KEYS[2]) end
                    return 1
                else
                    return nil
                end
                """
            )

        self.last_job = now()
        job.queue = self
        job.queued = now()
        job.status = JobStatus.QUEUED

        await self._before_enqueue(job)

        async with self._op_sem:
            if not await self._enqueue_script(
                    keys=[job.incomplete_key, job.id, job.queued_key, job.abort_id],
                    args=[self.serialize(job), job.scheduled],
                    client = self.ctx.async_client,
            ):
                return None
        if not self.client_mode and self.debug_enabled:
            self.logger(job=job, kind="enqueue").info(f"Enqueuing {job}")
        elif self.is_silenced_function(job.function, stage = 'enqueue'):
            pass
        # elif job.function in self.silenced_functions:
        #     pass
        elif self.verbose_results:
            self.logger(job=job, kind="enqueue").info(
                f"→ duration={now() - job.queued}ms, node={self.node_name}, func={job.function}, timeout={job.timeout}, kwargs={job.kwargs}"
            )
        elif self.truncate_logs:
            job_kwargs = f"{str({k: str(v)[:self.logging_max_length] for k, v in job.kwargs.items()})[:self.logging_max_length]}..." if job.kwargs else ""
            self.logger(job=job, kind="enqueue").info(
                f"→ duration={now() - job.queued}ms, node={self.node_name}, func={job.function}, timeout={job.timeout}, kwargs={job_kwargs}"
            )
        else:
            self.logger(job=job, kind="enqueue").info(
                f"→ duration={now() - job.queued}ms, node={self.node_name}, func={job.function}, timeout={job.timeout}"
            )

        return job

    async def apply(
        self, 
        job_or_func: typing.Union[Job, str],
        timeout: int = None, 
        broadcast: typing.Optional[bool] = None,
        worker_names: typing.Optional[typing.List[str]] = None,
        worker_selector: typing.Optional[typing.Callable] = None,
        worker_selector_args: typing.Optional[typing.List] = None,
        worker_selector_kwargs: typing.Optional[typing.Dict] = None,
        workers_selected: typing.Optional[typing.List[typing.Dict[str, str]]] = None,
        return_all_results: typing.Optional[bool] = False,
        **kwargs
    ):
        """
        Enqueue a job and wait for its result.

        If the job is successful, this returns its result.
        If the job is unsuccessful, this raises a JobError.

        Example::
            try:
                assert await queue.apply("add", a=1, b=2) == 3
            except JobError:
                print("job failed")

        job_or_func: Same as Queue.enqueue
        kwargs: Same as Queue.enqueue
        timeout: How long to wait for the job to complete before raising a TimeoutError
        broadcast: Broadcast the job to all workers. If True, the job will be run on all workers.
        worker_names: List of worker names to run the job on. If provided, will run on these specified workers.
        worker_selector: Function that takes in a list of workers and returns a list of workers to run the job on. If provided, worker_names will be ignored.
        """
        results = await self.map(
            job_or_func, 
            timeout = timeout, 
            iter_kwargs = [kwargs],
            broadcast = broadcast,
            worker_names = worker_names,
            worker_selector = worker_selector,
            worker_selector_args = worker_selector_args,
            worker_selector_kwargs = worker_selector_kwargs,
            workers_selected = workers_selected,
        )
        return results if return_all_results else results[0]
    
    async def select_workers(
        self,
        func: typing.Callable,
        *args,
        **kwargs,
    ) -> typing.List[typing.Dict[str, str]]:
        """
        Passes the dict of the worker attributes to the 
        provided function and returns a list of workers that
        are able to run the job.

        Function should expect kw: `worker_attributes`: Dict[str, Dict[str, Any]]
        and should return List[Dict[str, str]] where the Dict is either
            {'worker_id': str}
        or
            {'worker_name': str}
        """
        await self.prepare_for_broadcast()
        func = ensure_coroutine_function(func)
        return await func(
            *args,
            worker_attributes = self.worker_attributes,
            **kwargs
        )

    async def prepare_job_kws(
        self,
        iter_kwargs: typing.Iterable[typing.Dict], 
        timeout: int = None,
        broadcast: typing.Optional[bool] = False,
        worker_names: typing.Optional[typing.List[str]] = None,
        worker_selector: typing.Optional[typing.Callable] = None,
        worker_selector_args: typing.Optional[typing.List] = None,
        worker_selector_kwargs: typing.Optional[typing.Dict] = None,
        workers_selected: typing.Optional[typing.List[typing.Dict[str, str]]] = None,
        **kwargs
    ) -> typing.Tuple[typing.List[typing.Dict[str, typing.Any]], typing.Optional[typing.List[typing.Dict[str, str]]]]:
        """
        Prepare jobs for broadcast or map.

        Returns:
            iter_kwargs: List of kwargs to pass to the job
            worker_kws: List of worker kwargs to pass to the job
        """
        if (
            not broadcast
            and not worker_selector
            and not worker_names
            and not workers_selected
        ):
            return [
                {
                    "timeout": timeout,
                    "key": kwargs.get("key") or get_default_job_key(),
                    **kwargs,
                    **kw,
                }
                for kw in iter_kwargs
            ], None
        if workers_selected: worker_kws = workers_selected
        elif worker_selector:
            worker_selector_args = worker_selector_args or []
            worker_selector_kwargs = worker_selector_kwargs or {}
            worker_kws = await self.select_workers(worker_selector, *worker_selector_args, **worker_selector_kwargs)
        elif worker_names: worker_kws = [{"worker_name": worker_name} for worker_name in worker_names]

        else:
            await self.prepare_for_broadcast()
            # worker_kws = [{"worker_id": worker_id} for worker_id in self.active_workers]
            worker_kws = [{"worker_id": worker_id} for worker_id in await self.get_worker_ids()]

        broadcast_kwargs = []
        for kw in iter_kwargs:
            broadcast_kwargs.extend(
                {
                    "timeout": timeout,
                    "key": get_default_job_key(),
                    **kwargs,
                    **kw,
                    **worker_kw,
                }
                for worker_kw in worker_kws
            )
        return broadcast_kwargs, worker_kws

    async def broadcast(
        self,
        job_or_func: typing.Union[Job, str],
        timeout: int = None,
        enqueue: typing.Optional[bool] = True,
        worker_names: typing.Optional[typing.List[str]] = None,
        worker_selector: typing.Optional[typing.Callable] = None,
        worker_selector_args: typing.Optional[typing.List] = None,
        worker_selector_kwargs: typing.Optional[typing.Dict] = None,
        workers_selected: typing.Optional[typing.List[typing.Dict[str, str]]] = None,
        **kwargs
    ):
        """
        Broadcast a job to all nodes and collect all of their results.
        
        job_or_func: Same as Queue.enqueue
        kwargs: Same as Queue.enqueue
        timeout: How long to wait for the job to complete before raising a TimeoutError
        worker_names: List of worker names to run the job on. If provided, will run on these specified workers.
        worker_selector: Function that takes in a list of workers and returns a list of workers to run the job on. If provided, worker_names will be ignored.
        """

        if not enqueue:
            return await self.map(
                job_or_func, 
                timeout = timeout, 
                iter_kwargs = [kwargs], 
                broadcast = True, 
                worker_names = worker_names,
                worker_selector = worker_selector,
                worker_selector_args = worker_selector_args,
                worker_selector_kwargs = worker_selector_kwargs,
                workers_selected = workers_selected,
            )
    
        iter_kwargs, worker_kws = await self.prepare_job_kws(
            iter_kwargs = [kwargs], 
            timeout = timeout, 
            broadcast = True, 
            worker_names = worker_names, 
            worker_selector = worker_selector,
            worker_selector_args = worker_selector_args,
            worker_selector_kwargs = worker_selector_kwargs,
            workers_selected = workers_selected,
            **kwargs
        )
        # if worker_kws:
        #     # logger.info(f"Enqueueing {job_or_func} to {len(worker_kws)} workers: {worker_kws}: {iter_kwargs}")
        jobs: typing.List[Job] = []
        for kw in iter_kwargs:
            jobs.append(
                await self.enqueue(
                    job_or_func,
                    **kw
                )
            )
        return jobs



    async def map(
        self, 
        job_or_func: typing.Union[Job, str],
        iter_kwargs: typing.Iterable[typing.Dict], 
        timeout: int = None, 
        return_exceptions: bool = False, 
        broadcast: typing.Optional[bool] = False,
        worker_names: typing.Optional[typing.List[str]] = None,
        worker_selector: typing.Optional[typing.Callable] = None,
        worker_selector_args: typing.Optional[typing.List] = None,
        worker_selector_kwargs: typing.Optional[typing.Dict] = None,
        workers_selected: typing.Optional[typing.List[typing.Dict[str, str]]] = None,
        **kwargs
    ):
        """
        Enqueue multiple jobs and collect all of their results.

        Example::
            try:
                assert await queue.map(
                    "add",
                    [
                        {"a": 1, "b": 2},
                        {"a": 3, "b": 4},
                    ]
                ) == [3, 7]
            except JobError:
                print("any of the jobs failed")

        job_or_func: Same as Queue.enqueue
        iter_kwargs: Enqueue a job for each item in this sequence. Each item is the same
            as kwargs for Queue.enqueue.
        timeout: Total seconds to wait for all jobs to complete. If None (default) or 0, wait forever.
        return_exceptions: If False (default), an exception is immediately raised as soon as any jobs
            fail. Other jobs won't be cancelled and will continue to run.
            If True, exceptions are treated the same as successful results and aggregated in the result list.
        broadcast: If True, broadcast the job to all nodes. Otherwise, only enqueue the job on this node.
        kwargs: Default kwargs for all jobs. These will be overridden by those in iter_kwargs.
        """
        iter_kwargs, worker_kws = await self.prepare_job_kws(
            iter_kwargs = iter_kwargs, 
            timeout = timeout, 
            broadcast = broadcast, 
            worker_names = worker_names, 
            worker_selector = worker_selector, 
            workers_selected = workers_selected,
            worker_selector_args = worker_selector_args,
            worker_selector_kwargs = worker_selector_kwargs,
            **kwargs
        )
        if worker_kws and (self.verbose_results or self.debug_enabled):
            logger.info(f"Broadcasting {job_or_func} to {len(worker_kws)} workers: {worker_kws}")
            
        job_keys = [key["key"] for key in iter_kwargs]
        pending_job_keys = set(job_keys)
        async def callback(job_key, status):
            if status in TERMINAL_STATUSES:
                pending_job_keys.discard(job_key)

            if status in UNSUCCESSFUL_TERMINAL_STATUSES and not return_exceptions:
                return True

            if not pending_job_keys:
                return True

        # Start listening before we enqueue the jobs.
        # This ensures we don't miss any updates.
        task = asyncio.create_task(
            self.listen(pending_job_keys, callback, timeout=None)
        )

        try:
            await asyncio.gather(
                *[self.enqueue(job_or_func, **kw) for kw in iter_kwargs]
            )
        except:
            task.cancel()
            raise

        await asyncio.wait_for(task, timeout=timeout)

        jobs: typing.List[Job] = await asyncio.gather(*[self.job(job_key) for job_key in job_keys])

        results = []
        for job, job_key in zip(jobs, job_keys):
            if job is None:
                if not return_exceptions: raise ValueError(f"Job {job_key} not found")
                results.append(None)
            elif job.status in UNSUCCESSFUL_TERMINAL_STATUSES:
                exc = JobError(job)
                if not return_exceptions: raise exc
                results.append(exc)
            else: results.append(job.result)
        return results


    """
    Batch Methods
    """

    
    async def batch_map(
        self, 
        batch_func_kwargs: typing.Union[typing.List, typing.Dict],
        timeout: int = None, 
        return_exceptions: bool = False, 
        **kwargs
    ):
        """
        Enqueue multiple individual as a result of multiple funcs collect all of their results.

        Example::
            try:
                assert await queue.batch_map(
                    [
                        "add", {"a": 1, "b": 2},
                        "subtract", {"a": 3, "b": 2},
                    ]
                ) == [3, 1]
            except JobError:
                print("any of the jobs failed")
            
            try:
                assert await queue.batch_map(
                    {
                        "result_1": ["add", {"a": 1, "b": 2}],
                        "result_2": ["subtract", {"a": 3, "b": 2}],
                    }
                ) == {"result_1": 3, "result_2": 1}
            except JobError:
                print("any of the jobs failed")

        job_or_func: Same as Queue.enqueue
        iter_kwargs: Enqueue a job for each item in this sequence. Each item is the same
            as kwargs for Queue.enqueue.
        timeout: Total seconds to wait for all jobs to complete. If None (default) or 0, wait forever.
        return_exceptions: If False (default), an exception is immediately raised as soon as any jobs
            fail. Other jobs won't be cancelled and will continue to run.
            If True, exceptions are treated the same as successful results and aggregated in the result list.
        kwargs: Default kwargs for all jobs. These will be overridden by those in iter_kwargs.
        """
        if isinstance(batch_func_kwargs, dict):
            return await self.batch_map_dict(
                batch_func_kwargs, timeout = timeout, return_exceptions = return_exceptions, **kwargs
            )
        return await self.batch_map_list(
            batch_func_kwargs, timeout = timeout, return_exceptions = return_exceptions, **kwargs
        )

    async def batch_map_dict(
        self, 
        batch_func_kwargs: typing.Dict[str, typing.Tuple[str, typing.Dict]], 
        timeout: int = None, 
        return_exceptions: bool = False, 
        **kwargs
    ):
        """
        Enqueue multiple individual as a result of multiple funcs collect all of their results.

        Example::
            try:
                assert await queue.batch_map(
                    {
                        "result_1": ("add", {"a": 1, "b": 2}),
                        "result_2": ("subtract", {"a": 3, "b": 2}),
                    }
                ) == {"result_1": 3, "result_2": 1}
            except JobError:
                print("any of the jobs failed")

        batch_func_kwargs: Same as Queue.enqueue
        iter_kwargs: Enqueue a job for each item in this sequence. Each item is the same
            as kwargs for Queue.enqueue.
        timeout: Total seconds to wait for all jobs to complete. If None (default) or 0, wait forever.
        return_exceptions: If False (default), an exception is immediately raised as soon as any jobs
            fail. Other jobs won't be cancelled and will continue to run.
            If True, exceptions are treated the same as successful results and aggregated in the result list.
        kwargs: Default kwargs for all jobs. These will be overridden by those in iter_kwargs.
        """
        batch_kwargs = {
            key: {
                "job_or_func": func,
                "timeout": timeout,
                "key": kw.get("key") or get_default_job_key(),
                **kwargs,
                **kw,
            }
            for key, (func, kw) in batch_func_kwargs.items()
        }
        job_dict = {kw["key"]: key  for key, kw in batch_kwargs.items()}
        pending_job_keys = set(job_dict.keys())

        async def callback(job_key, status):
            if status in TERMINAL_STATUSES: pending_job_keys.discard(job_key)
            if status in UNSUCCESSFUL_TERMINAL_STATUSES and not return_exceptions: return True
            if not pending_job_keys: return True

        # Start listening before we enqueue the jobs.
        # This ensures we don't miss any updates.
        task = asyncio.create_task(self.listen(pending_job_keys, callback, timeout=None))

        try:
            await asyncio.gather(
                *[self.enqueue(**kw) for kw in list(batch_kwargs.values())]
            )
        except:
            task.cancel()
            raise

        await asyncio.wait_for(task, timeout=timeout)
        jobs: typing.List[Job] = await asyncio.gather(*[self.job(job_key) for job_key in job_dict])
        results = {}
        
        for job, job_key in zip(jobs, job_dict):
            if job is None:
                if not return_exceptions: raise ValueError(f"Job {job_key} not found")
                results[job_dict[job_key]] = None
            elif job.status in UNSUCCESSFUL_TERMINAL_STATUSES:
                exc = JobError(job)
                if not return_exceptions: raise exc
                results[job_dict[job.key]] = exc
            else: results[job_dict[job.key]] = job.result
        return results

    async def batch_map_list(
        self, 
        batch_func_kwargs: typing.List[typing.Tuple[str, typing.Dict[str, typing.Any]]], 
        timeout: int = None, 
        return_exceptions: bool = False, 
        **kwargs
    ):
        """
        Enqueue multiple individual as a result of multiple funcs collect all of their results.

        Example::
            try:
                assert await queue.batch_map(
                    [
                        ("add", {"a": 1, "b": 2}),
                        ("subtract", {"a": 3, "b": 2}),
                    ]
                ) == [3, 1]
            except JobError:
                print("any of the jobs failed")

        batch_func_kwargs: Same as Queue.enqueue
        iter_kwargs: Enqueue a job for each item in this sequence. Each item is the same
            as kwargs for Queue.enqueue.
        timeout: Total seconds to wait for all jobs to complete. If None (default) or 0, wait forever.
        return_exceptions: If False (default), an exception is immediately raised as soon as any jobs
            fail. Other jobs won't be cancelled and will continue to run.
            If True, exceptions are treated the same as successful results and aggregated in the result list.
        kwargs: Default kwargs for all jobs. These will be overridden by those in iter_kwargs.
        """
        batch_kwargs = [
            {
                "job_or_func": func,
                "timeout": timeout,
                "key": kw.get("key") or get_default_job_key(),
                **kwargs,
                **kw,
            }
            for (func, kw) in batch_func_kwargs
        ]
        job_keys = [key["key"] for key in batch_kwargs]
        pending_job_keys = set(job_keys)

        async def callback(job_key, status):
            if status in TERMINAL_STATUSES: pending_job_keys.discard(job_key)
            if status in UNSUCCESSFUL_TERMINAL_STATUSES and not return_exceptions: return True
            if not pending_job_keys: return True

        # Start listening before we enqueue the jobs.
        # This ensures we don't miss any updates.
        task = asyncio.create_task(self.listen(pending_job_keys, callback, timeout=None))

        try:
            await asyncio.gather(
                *[self.enqueue(**kw) for kw in batch_kwargs]
            )
        except:
            task.cancel()
            raise

        await asyncio.wait_for(task, timeout=timeout)
        jobs: typing.List[Job] = await asyncio.gather(*[self.job(job_key) for job_key in job_keys])
        results = []
        for job, job_key in zip(jobs, job_keys):
            if job is None:
                if not return_exceptions: raise ValueError(f"Job {job_key} not found")
                results.append(None)
            elif job.status in UNSUCCESSFUL_TERMINAL_STATUSES:
                exc = JobError(job)
                if not return_exceptions: raise exc
                results.append(exc)
            else: results.append(job.result)
        return results

    @asynccontextmanager
    async def batch(self):
        """
        Context manager to batch enqueue jobs.

        This tracks all jobs enqueued within the context manager scope and ensures that
        all are aborted if any exception is raised.

        Example::
            async with queue.batch():
                await queue.enqueue("test")  # This will get cancelled
                raise asyncio.CancelledError
        """
        children = set()

        async def track_child(job):
            children.add(job)

        self.register_before_enqueue(track_child)
        try: yield
        except:
            await asyncio.gather(
                *[self.abort(child, "cancelled") for child in children],
                return_exceptions=True,
            )
            raise
        finally: self.unregister_before_enqueue(track_child)

    """
    New Helper Methods
    """

    async def wait_for_jobs(
        self,
        jobs: typing.List[Job],
        source_job: typing.Optional[Job] = None,
        verbose: typing.Optional[bool] = False,
        raise_exceptions: typing.Optional[bool] = False,
        refresh_interval: typing.Optional[float] = 0.5,
        **kwargs,
    ) -> typing.List[typing.Any]:
        """
        Waits for jobs to finish
        """
        results, num_jobs = [], len(jobs)
        t = timer()
        while jobs:
            for job in jobs:
                try: await job.refresh()
                except RuntimeError: await job.enqueue()
                if job.status == JobStatus.COMPLETE:
                    results.append(job.result)
                    if source_job: await source_job.incr_progress()
                    jobs.remove(job)
                elif job.status == JobStatus.FAILED:
                    if verbose: logger.error(f'Job {job.id} failed: {job.error}')
                    if raise_exceptions: raise job.error
                    jobs.remove(job)
                if not jobs: break
            await asyncio.sleep(refresh_interval)
        
        if verbose: 
            if source_job: logger.info(f'Completed {source_job.id} w/ {len(results)}/{num_jobs} in {timer(t):.2f}s')
            else: logger.info(f'Completed {len(results)}/{num_jobs} jobs in {timer(t):.2f}s')
        return results        

    
    async def as_jobs_complete(
        self,
        jobs: typing.List[Job],
        source_job: typing.Optional[Job] = None,
        verbose: typing.Optional[bool] = False,
        raise_exceptions: typing.Optional[bool] = False,
        refresh_interval: typing.Optional[float] = 0.5,
        return_results: typing.Optional[bool] = True,
        **kwargs,
    ) -> typing.AsyncGenerator[typing.Any, None]:
        # sourcery skip: low-code-quality
        """
        Generator that yields results as they complete
        """
        t = timer()
        num_results, num_jobs = 0, len(jobs)
        while jobs:
            for job in jobs:
                try: await job.refresh()
                except RuntimeError: await job.enqueue()
                if job.status == JobStatus.COMPLETE:
                    yield job.result if return_results else job
                    num_results += 1
                    if source_job: await source_job.incr_progress()
                    jobs.remove(job)
                
                elif job.status == JobStatus.FAILED:
                    if verbose: logger.error(f'Job {job.id} failed: {job.error}')
                    if raise_exceptions: raise job.error
                    jobs.remove(job)
                
                if not jobs: break
            await asyncio.sleep(refresh_interval)
        
        if verbose: 
            if source_job: logger.info(f'Completed {source_job.id} w/ {num_results}/{num_jobs} in {timer(t):.2f}s')
            else: logger.info(f'Completed {num_results}/{num_jobs} jobs in {timer(t):.2f}s')

    @property
    def batch_iterator_concurrency_limit(self) -> int:
        """
        Returns the batch iterator concurrency limit
        """
        return self.max_concurrency - 5

    async def batch_enqueue_iterator(
        self,
        job_or_func: typing.Union[Job, str],
        iterable: typing.Iterable[typing.Any],
        iterable_key: typing.Optional[str] = None,
        concurrency_limit: typing.Optional[int] = None,
        enqueue_func: typing.Optional[typing.Callable] = None,
        _verbose: typing.Optional[bool] = False,
        _raise_exceptions: typing.Optional[bool] = False,
        _return_results: typing.Optional[bool] = True,
        _key_func: typing.Optional[typing.Callable] = None,
        **kwargs,
    ) -> typing.AsyncGenerator[typing.Any, None]:
        """
        Helper method to batch enqueue jobs with a concurrency limit.
        and yields the results as they complete.
        """
        enqueue_func = enqueue_func or self.enqueue
        concurrency_limit = concurrency_limit or self.max_concurrency - 5
        n_items, t = 0, timer()
        for batch in build_batches(iterable, concurrency_limit, fixed_batch_size = True):
            batch_jobs = []
            bt = timer()
            for item in batch:
                if _key_func: kwargs['key'] = _key_func(item)
                if iterable_key: kwargs[iterable_key] = item
                batch_jobs.append(
                    await enqueue_func(
                        job_or_func,
                        *(() if iterable_key else (item)),
                        **kwargs,
                    )
                )
            async for completed_job in self.as_jobs_complete(
                jobs = batch_jobs,
                verbose = _verbose,
                raise_exceptions = _raise_exceptions,
                return_results = _return_results,
            ):
                yield completed_job
                n_items += 1
            
            if _verbose: logger.info(f'[{job_or_func}] Completed {len(batch_jobs)} Batch Results in {timer(bt):.2f}s ({len(batch_jobs) / timer(bt):.2f} items/s)')
        if _verbose: logger.info(f'[{job_or_func}] Completed {n_items} Results in {timer(t):.2f}s ({n_items / timer(t):.2f} items/s)')


    @property
    async def _all_queue_job_keys(self) -> typing.List[str]:
        """
        Returns all job ids in the queue
        """
        keys = await self.ctx.async_keys(f'queue:{self.queue_name}:job:*')
        return [key.decode('utf-8') for key in keys]

    @property
    async def _all_queue_job_ids(self) -> typing.List[str]:
        """
        Returns all job ids in the queue
        """
        keys = await self._all_queue_job_keys
        return [key.split(':')[-1] for key in keys]
    
    @property
    async def _all_queue_function_tracker_keys(self) -> typing.List[str]:
        """
        Returns all function tracker keys in the queue
        """
        keys = await self.ctx.async_keys(f'queue:{self.queue_name}:functiontracker.*')
        return [key.decode('utf-8') for key in keys]

    async def _reset_queue_jobs(
        self, 
        error: typing.Optional[typing.Union[str, Exception]] = 'Resetting Queue',
        ttl: typing.Optional[int] = 5,
        status: typing.Optional[JobStatus] = JobStatus.ABORTED,
        verbose: typing.Optional[bool] = True,
        **kwargs,
    ):
        """
        Helper method to reset all jobs in the queue
        """
        job_keys = await self._all_queue_job_keys
        for job_key in job_keys:
            job = await self.job(job_key)
            try:
                await self.abort(job, error = error, ttl = ttl)
                await self.ctx.async_client.publish(job.id, status)
            except Exception as e:
                logger.error(f'Error resetting job {job_key}: {e}')
        if verbose: logger.info(f'Reset {len(job_keys)} jobs in {self.queue_name}')

    async def _reset_queue_trackers(
        self,
        verbose: typing.Optional[bool] = True,
        **kwargs,
    ):
        """
        Helper method to reset all function trackers in the queue
        """
        tracker_keys = await self._all_queue_function_tracker_keys
        for tracker_key in tracker_keys:
            await self.ctx.async_delete(tracker_key)
        if verbose: logger.info(f'Reset {len(tracker_keys)} function trackers in {self.queue_name}')

    async def reset_queue(
        self,
        jobs: typing.Optional[bool] = False,
        # scheduled: typing.Optional[bool] = False,
        tracker: typing.Optional[bool] = False,
        verbose: typing.Optional[bool] = True,
        **kwargs,
    ):
        """
        Purges all keys from the queue
        """
        if jobs: await self._reset_queue_jobs(verbose = verbose, **kwargs)
        if tracker: await self._reset_queue_trackers(verbose = verbose, **kwargs)


    """
    Queue Metadata
    """
    
    async def fetch_job_info(
        self,
        offset: int = 0, 
        limit: int = 10,
        **kwargs
    ) -> typing.List[typing.Union[Job, typing.Dict]]:
        """
        Fetch all job info
        """
        job_info = []
        for job_bytes in await self.ctx.async_mget(
            (
                await self.ctx.async_lrange(self.active_key, offset, limit - 1)
            )
            + (
                await self.ctx.async_lrange(self.queued_key, offset, limit - 1)
            )
        ):
            with suppress(Exception):
                job_info.append(self.deserialize(job_bytes).to_dict())
        return job_info

    async def get_worker_ids(self) -> typing.List[str]:
        """
        Get all worker ids
        """
        worker_ids = []
        for key in await self.ctx.async_zrangebyscore(self.heartbeat_key, now(), "inf"):
            key: bytes = key.decode("utf-8")
            *_, worker_uuid = key.split(":")
            worker_ids.append(worker_uuid)
        return worker_ids

    async def info(self, jobs: bool = False, offset: int = 0, limit: int = 10):
        # pylint: disable=too-many-locals
        worker_uuids = []

        for key in await self.ctx.async_zrangebyscore(self.stats_key, now(), "inf"):
            key: bytes = key.decode("utf-8")
            *_, worker_uuid = key.split(":")
            worker_uuids.append(worker_uuid)

        worker_stats = await self.ctx.async_mget(
            self.create_namespace(f"stats:{worker_uuid}") for worker_uuid in worker_uuids
        )
        worker_info = {}
        for worker_uuid, stats in zip(worker_uuids, worker_stats):
            if stats:
                stats = json.loads(stats.decode("UTF-8"))
                worker_info[worker_uuid] = stats
        

        metrics = await self.prepare_queue_metrics(worker_info)
        job_info = await self.fetch_job_info(offset = offset, limit = limit) if jobs else []
        metrics['jobs'] = job_info
        return metrics

    async def sync_queue_info(self):
        """
        Syncs the current queue info with keydb
        """
        queue_info = await self.info(jobs = True)
        await self.ctx.async_set(self.queue_info_key, json.dumps(queue_info, cls = ObjectEncoder), ex = 60)
    
    async def fetch_queue_info(self):
        """
        Fetches the current queue info from keydb
        """
        queue_info = await self.ctx.async_get(self.queue_info_key)
        return json.loads(queue_info) if queue_info else {}

    async def stats(self, ttl: int = 60):
        """
        Logs the current queue stats to keydb
        """
        try:
            async with anyio.fail_after(10):
                stats = await self.prepare_worker_metrics()
                current = now()
                # async with self.pipeline(transaction = True) as pipe:
                async with self.ctx.async_client.pipeline(transaction = True, retryable = True) as pipe:
                    key = self.create_namespace(f"stats:{self.uuid}")
                    await (
                        pipe.setex(key, ttl, json.dumps(stats))
                        .zremrangebyscore(self.stats_key, 0, current)
                        .zadd(self.stats_key, {key: current + millis(ttl)})
                        .expire(self.stats_key, ttl)
                        .execute()
                    )
                return stats
        except TimeoutError as e:
            logger.error(f"TimeoutError while logging stats: {e}")
        except Exception as e:
            logger.error(f"Error while logging stats: {e}")

    
    async def add_heartbeat(
        self, 
        worker_id: str, 
        worker_attributes: typing.Dict[str, typing.Any] = None,
        heartbeat_ttl: typing.Optional[int] = None,
    ):
        """
        Registers a heartbeat for the current worker.
        """
        current = now()
        worker_attributes = worker_attributes or {}
        heartbeat_ttl = heartbeat_ttl or self.heartbeat_ttl
        async with self.ctx.async_client.pipeline(transaction = True, retryable = True) as pipe:
        # async with self.pipeline(transaction = True) as pipe:
            key = self.create_namespace(f"worker:attr:{worker_id}")
            await (
                pipe.setex(key, heartbeat_ttl, json.dumps(worker_attributes))
                .zremrangebyscore(self.heartbeat_key, 0, current)
                .zadd(self.heartbeat_key, {worker_id: current + millis(heartbeat_ttl)})
                .expire(self.heartbeat_key, heartbeat_ttl)
                .execute()
            )
        self.active_workers = await self.get_worker_ids()
        self.num_workers = len(self.active_workers)


    async def count(self, kind: str, postfix: typing.Optional[str] = None):
        """
        Returns the number of jobs in the given queue.
        """
        if kind == "queued":
            return await self.ctx.async_llen(f"{self.queued_key}:{postfix}" if postfix else self.queued_key)
        if kind == "active":
            return await self.ctx.async_llen(f"{self.active_key}:{postfix}" if postfix else self.active_key)
        if kind == "scheduled":
            return await self.ctx.async_zcount(f"{self.scheduled_key}:{postfix}" if postfix else self.scheduled_key, 1, "inf")
        if kind == "incomplete":
            return await self.ctx.async_zcard(f"{self.incomplete_key}:{postfix}" if postfix else self.incomplete_key)
        raise ValueError(f"Can't count unknown type {kind} {postfix}")

    @property
    def _queue_kind_key_map(self) -> typing.Dict[str, str]:
        """
        Returns the mapping of queue kind to key
        """
        return {
            "queued": self.queued_key,
            "active": self.active_key,
            "scheduled": self.scheduled_key,
            "incomplete": self.incomplete_key,
        }

    @property
    async def _all_queue_jobs(self) -> typing.Dict[str, JobStatus]:
        """
        Returns all jobs in all queues
        """
        job_ids = await self.ctx.async_hgetall(self._stats.queue_job_ids_key)
        if not job_ids: return {}
        return {
            job_id.decode("utf-8"): JobStatus(job_status.decode("utf-8")) for job_id, job_status in job_ids.items()
        } 

    async def get_job_ids_in_queue(self, kind: typing.Optional[str] = None, key: typing.Optional[str] = None, postfix: typing.Optional[str] = None, encode: typing.Optional[bool] = True) -> typing.List[str]:
        """
        Returns all jobs in the given queue.
        """
        if not kind:
            job_ids = await self.ctx.async_hgetall(self._stats.queue_job_ids_key)
            # logger.info(f"Job ids: {job_ids}")
            return [job_id.decode("utf-8") for job_id in job_ids.keys()] if encode else job_ids.keys()

        if not key:
            base_key = self._queue_kind_key_map[kind]
            key = f"{base_key}:{postfix}" if postfix else base_key
        results: typing.List[bytes] = await self.ctx.async_lrange(key, 0, -1)
        return [result.decode("utf-8") for result in results] if encode else results
    
    async def sweep_jobs_by_status(self, status: JobStatus, kind: typing.Optional[str] = "active", key: typing.Optional[str] = None, postfix: typing.Optional[str] = None, verbose: typing.Optional[bool] = None, **kwargs):
        """
        Sweeps all jobs with the given status.
        """
        verbose = verbose or self._should_debug_log
        job_ids = await self.get_job_ids_in_queue(kind, key = key, postfix = postfix)
        if not job_ids: 
            if verbose: logger.info(f"No jobs found in {kind} queue")
            return
        if verbose: logger.info(f"Sweeping {len(job_ids)} jobs from {kind} queue with status {status}")

        async def _sweep_job(job_id: str):
            """
            Inner sweep job function
            """
            job = await self._get_job_by_id(job_id)
            if job.status == status:
                if verbose: logger.info(f"Sweeping job {job_id} with status {status}")
                await self.sweep_job(job = job, job_id = job_id, verbose = verbose)
            else:
                if verbose: logger.info(f"Skipping job {job_id} with status {job.status}")

        async for _ in concurrent_map(
            _sweep_job,
            job_ids,
        ):
            pass
        return job_ids


    """
    Metrics Handler
    """
    async def prepare_for_broadcast(self, ttl: typing.Optional[int] = None):
        """
        Fetches the current workers
        """
        worker_data = await self.ctx.async_mget(
            self.create_namespace(f"worker:attr:{worker_id}") for worker_id in self.active_workers
        )
        self.worker_attributes = {
            worker_id: json.loads(worker_attr.decode("UTF-8"))
            for worker_id, worker_attr in zip(self.active_workers, worker_data)
            if worker_attr
        }
        

    async def prepare_queue_metrics(
        self, 
        worker_info: typing.Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs
    ) -> typing.Dict[str, typing.Any]:
        """
        Parse worker metrics from worker info.
        """
        return worker_info
    
    async def prepare_worker_metrics(
        self, 
        **kwargs
    ) -> typing.Dict:
        """
        Parse worker metrics from worker info.
        """
        pass

    """
    Function Tracker
    """

    async def setup_function_tracker(self, functions: typing.List[str]):
        """
        Sets up the function tracker for the given functions.
        """
        funcs = {}
        for func in functions:
            _existing = await self.ctx.async_get(f'{self._stats.function_tracker_key}.{func}', default = None)
            funcs[func] = FunctionTracker.deserialize(_existing) if _existing else FunctionTracker(function = func)
        self._function_tracker = funcs
        logger.info(f"Function Tracker: {self._function_tracker}")

    @asynccontextmanager
    async def _fail_ok(self, duration: typing.Optional[float] = 7.0):
        """
        Allows a block of code to fail without raising an exception.
        """
        try:
            async with anyio.fail_after(duration):
                yield
        except Exception as e:
            logger.trace(f"Failed OK: {e}", error = e, level = "WARNING")

    async def _get_function_tracker(self, function: str, none_ok: typing.Optional[bool] = True) -> typing.Optional[FunctionTracker]:
        """
        Returns the function tracker for the given function.
        """
        if not self.function_tracker_enabled: return None
        function_tracker = await self.ctx.async_get(f'{self._stats.function_tracker_key}.{function}', default = None)
        if function_tracker:
            function_tracker = FunctionTracker.deserialize(function_tracker)
        elif none_ok: return None
        else:
            function_tracker = FunctionTracker(function = function)
        return function_tracker
    
    async def track_job_id(self, job: Job):
        """
        Inserts the job id into the queue
        """
        async with self._fail_ok():
            if job.status in TERMINAL_STATUSES:
                await self.ctx.async_hdel(self._stats.queue_job_ids_key, job.id)
                # logger.info(f"Dropping job id {job.id}")
            else:
                await self.ctx.async_hset(self._stats.queue_job_ids_key, job.id, job.status.value)
                # logger.info(f"Inserting job id {job.id} with status {job.status.value}")
                # logger.info(f"Job ids: {await self.ctx.async_hgetall(self._stats.queue_job_ids_key)}")

    async def track_job(self, job: Job):
        """
        Tracks the job results
        """
        if not self.function_tracker_enabled: return
        async with self._fail_ok():
            function_tracker = await self._get_function_tracker(job.function, none_ok = False)
            function_tracker.track_job(job)
            await self.ctx.async_set(f'{self._stats.function_tracker_key}.{job.function}', function_tracker.serialize(), ex = self.function_tracker_ttl)
            await self.track_job_id(job)

    async def get_function_trackers(self) -> typing.Dict[str, FunctionTracker]:
        """
        Returns the function trackers
        """
        if not self.function_tracker_enabled: return {}
        _keys = await self.ctx.async_keys(f'{self._stats.function_tracker_key}.*')
        _function_trackers = {}
        for key in _keys:
            function_tracker = await self.ctx.async_get(key, default = None)
            if function_tracker:
                function_tracker = FunctionTracker.deserialize(function_tracker)
                _function_trackers[function_tracker.function] = function_tracker
        return _function_trackers
    
    async def reset_function_trackers(self, functions: typing.Optional[typing.List[str]] = None):
        """
        Resets the function trackers
        """
        if not self.function_tracker_enabled: return
        async with self._fail_ok():
            if functions is None: 
                functions = await self.ctx.async_keys(f'{self._stats.function_tracker_key}.*')
            else:
                functions = [f'{self._stats.function_tracker_key}.{function}' for function in functions]
            for function in functions:
                await self.ctx.async_delete(f'{self._stats.function_tracker_key}.{function}')

    """
    Management Methods
    """

    """
    Alias Properties
    """
    # Properties
    @lazyproperty
    def node_name(self) -> str: return self._stats.node_name

    
    # Counts
    @lazyproperty
    def complete(self) -> int: return self._stats.complete
    @lazyproperty
    def failed(self) -> int: return self._stats.failed
    @lazyproperty
    def retried(self) -> int: return self._stats.retried
    @lazyproperty
    def aborted(self) -> int: return self._stats.aborted
    @lazyproperty
    def num_workers(self) -> int: return self._stats.num_workers
    
    # Key Namespaces
    @lazyproperty
    def active_key(self) -> str: return self._stats.active_key
    @lazyproperty
    def queued_key(self) -> str: return self._stats.queued_key
    @lazyproperty
    def scheduled_key(self) -> str: return self._stats.scheduled_key
    @lazyproperty
    def incomplete_key(self) -> str: return self._stats.incomplete_key
    @lazyproperty
    def function_tracker_key(self) -> str: return self._stats.function_tracker_key
    @lazyproperty
    def stats_key(self) -> str: return self._stats.stats_key
    @lazyproperty
    def sweep_key(self) -> str: return self._stats.sweep_key
    @lazyproperty
    def queue_info_key(self) -> str: return self._stats.queue_info_key
    @lazyproperty
    def worker_token_key(self) -> str: return self._stats.worker_token_key
    @lazyproperty
    def management_queue_key(self) -> str: return self._stats.management_queue_key
    @lazyproperty
    def heartbeat_key(self) -> str: return self._stats.heartbeat_key
    @lazyproperty
    def abort_id_prefix(self) -> str: return self._stats.abort_id_prefix

    @lazyproperty
    def active_workers(self) -> typing.List[str]: return self._stats.active_workers
    @lazyproperty
    def worker_attributes(self) -> typing.Dict[str, typing.Dict[str, typing.Any]]: return self._stats.worker_attributes

    def create_namespace(self, key: str) -> str: return self._stats.create_namespace(key)

# Job.update_forward_refs(queue = TaskQueue)
# Job.update_forward_refs()
# Job.__try_update_forward_refs__()


from __future__ import annotations

import enum

import typing
import datetime
import croniter

from aiokeydb.v2.types.base import BaseModel, lazyproperty, Field, validator
from aiokeydb.v2.utils.queue import (
    get_default_job_key, 
    now, 
    seconds, 
    exponential_backoff,
    import_function,
    get_func_name,
    get_func_full_name,
    ensure_coroutine_function,
    create_background_task
)
from aiokeydb.v2.configs import settings
from aiokeydb.v2.utils.logs import logger
from aiokeydb.v2.types.static import JobStatus, TaskType, TERMINAL_STATUSES, UNSUCCESSFUL_TERMINAL_STATUSES


if typing.TYPE_CHECKING:
    from aiokeydb.v2.types.task_queue import TaskQueue



class CronJob(BaseModel):
    """
    Allows scheduling of repeated jobs with cron syntax.

    function: the async function to run
    cron: cron string for a job to be repeated, uses croniter
    unique: unique jobs only one once per queue, defaults true

    Remaining kwargs are pass through to Job
    """

    function: typing.Callable
    cron: str
    unique: bool = True
    cron_name: typing.Optional[str] = None
    timeout: typing.Optional[int] = Field(default_factory = settings.get_default_job_timeout)
    retries: typing.Optional[int] = Field(default_factory = settings.get_default_job_retries)
    ttl: typing.Optional[int] = Field(default_factory = settings.get_default_job_ttl)
    heartbeat: typing.Optional[int] = None
    default_kwargs: typing.Optional[dict] = None

    # Allow for passing in a callback function to be called after the job is run
    callback: typing.Optional[typing.Union[str, typing.Callable]] = None
    callback_kwargs: typing.Optional[dict] = Field(default_factory=dict)

    @property
    def function_name(self) -> str:
        """
        Returns the name of the function
        """
        return self.cron_name or self.function.__qualname__
    
    @validator("callback")
    def validate_callback(cls, v: typing.Optional[typing.Union[str, typing.Callable]]) -> typing.Optional[str]:
        """
        Validates the callback and returns the function name
        """
        return v if v is None else get_func_full_name(v)
    
    def next_scheduled(self) -> int:
        """
        Returns the next scheduled time for the cron job
        """
        return int(croniter.croniter(self.cron, seconds(now())).get_next())

    def to_enqueue_kwargs(self, job_key: typing.Optional[str] = None, exclude_none: typing.Optional[bool] = True, **kwargs) -> typing.Dict[str, typing.Any]:
        """
        Returns the kwargs for the job
        """
        default_kwargs = self.default_kwargs or {}
        if kwargs: default_kwargs.update(kwargs)
        default_kwargs['key'] = job_key
        enqueue_kwargs = {
            "function": self.function_name,
            **default_kwargs,
        }
        if self.callback:
            enqueue_kwargs['job_callback'] = self.callback
            enqueue_kwargs['job_callback_kwargs'] = self.callback_kwargs

        if exclude_none:
            enqueue_kwargs = {
                k: v
                for k, v in enqueue_kwargs.items()
                if v is not None
            }
        enqueue_kwargs['scheduled'] = self.next_scheduled()
        return enqueue_kwargs
    

class JobProgress(BaseModel):

    """
    Holds the progress of a job
    """

    total: int = 0
    completed: int = 0

    @property
    def progress(self) -> float:
        """
        Returns the progress of a job as a float between 0.0 and 1.0
        """
        return 0.0 if self.total == 0 else self.completed / self.total


class FunctionTracker(BaseModel):
    """
    Tracks the metrics for a job function
    """
    function: str
    completed_durations: typing.List[float] = Field(default_factory=list)
    failed_durations: typing.List[float] = Field(default_factory=list)
    last_completed: typing.Optional[datetime.datetime] = None
    last_failed: typing.Optional[datetime.datetime] = None

    @property
    def durations(self) -> typing.List[float]:
        """
        Returns the durations of a job function
        """
        return self.completed_durations + self.failed_durations
    
    @property
    def failed(self) -> int:
        """
        Returns the number of failed jobs
        """
        return len(self.failed_durations)
    
    @property
    def completed(self) -> int:
        """
        Returns the number of completed jobs
        """
        return len(self.completed_durations)

    @property
    def total_duration_ms(self) -> float:
        """
        Returns the total duration of a job function
        in milliseconds
        """
        return sum(self.durations)
    
    @property
    def total_duration(self) -> float:
        """
        Returns the total duration of a job function
        in seconds
        """
        return self.total_duration_ms / 1000.0
    
    @property
    def completed_duration_ms(self) -> float:
        """
        Returns the total duration of completed jobs
        in milliseconds
        """
        return sum(self.completed_durations)
    
    @property
    def completed_duration(self) -> float:
        """
        Returns the total duration of completed jobs
        in seconds
        """
        return self.completed_duration_ms / 1000.0
    
    @property
    def failed_duration_ms(self) -> float:
        """
        Returns the total duration of failed jobs
        in milliseconds
        """
        return sum(self.failed_durations)
    
    @property
    def failed_duration(self) -> float:
        """
        Returns the total duration of failed jobs
        in seconds
        """
        return self.failed_duration_ms / 1000.0
    
    @property
    def total(self) -> int:
        """
        Returns the total number of jobs
        """
        return self.completed + self.failed

    @property
    def average_duration_ms(self) -> float:
        """
        Returns the average duration of a job function
        in milliseconds
        """
        return 0.0 if self.completed == 0 else self.completed_duration_ms / self.completed
    
    @property
    def average_duration(self) -> float:
        """
        Returns the average duration of a job function
        in seconds
        """
        return self.average_duration_ms / 1000.0
    
    @property
    def success_rate(self) -> float:
        """
        Returns the success rate of a job function
        """
        return 0.0 if self.total == 0 else self.completed / self.total
    
    @property
    def failure_rate(self) -> float:
        """
        Returns the failure rate of a job function
        """
        return 0.0 if self.total == 0 else self.failed / self.total
    
    def track_job(self, job: 'Job'):
        """
        Tracks the job in the function tracker
        """
        if job.status in UNSUCCESSFUL_TERMINAL_STATUSES:
            self.failed_durations.append(job.job_duration)
            self.last_failed = datetime.datetime.now(tz = datetime.timezone.utc)
        elif job.status in TERMINAL_STATUSES:
            self.completed_durations.append(job.job_duration)
            self.last_completed = datetime.datetime.now(tz = datetime.timezone.utc)

    def serialize(self) -> str:
        """
        Serializes the function tracker
        """
        return self.json()
    
    @classmethod
    def deserialize(cls, data: str) -> 'FunctionTracker':
        """
        Deserializes the function tracker
        """
        return cls.parse_raw(data)
    
    @property
    def data_dict(self) -> dict:
        """
        Returns the function tracker as a dictionary
        """
        return {
            "function": self.function,
            "total": self.total,
            "completed": self.completed,
            "failed": self.failed,
            "total_duration": self.total_duration,
            "average_duration": self.average_duration,
            "success_rate": self.success_rate * 100,
            "failure_rate": self.failure_rate * 100,
            "last_completed": self.last_completed.isoformat() if self.last_completed else None,
            "last_failed": self.last_failed.isoformat() if self.last_failed else None,
        }



class Job(BaseModel):
    """
    Main job class representing a run of a function.

    User Provided Arguments
        function: the async function name to run
        kwargs: kwargs to pass to the function
        queue: the saq.Queue object associated with the job
        key: unique identifier of a job, defaults to uuid1, can be passed in to avoid duplicate jobs
        timeout: the maximum amount of time a job can run for in seconds, defaults to 600 (0 means disabled)
        heartbeat: the maximum amount of time a job can survive without a heartebat in seconds, defaults to 0 (disabled)
            a heartbeat can be triggered manually within a job by calling await job.update()
        retries: the maximum number of attempts to retry a job, defaults to 1
        ttl: the maximum time in seconds to store information about a job including results, defaults to 600 (0 means indefinitely, -1 means disabled)
        retry_delay: seconds to delay before retrying the job
        retry_backoff: If true, use exponential backoff for retry delays.
            The first retry will have whatever retry_delay is.
            The second retry will have retry_delay*2. The third retry will have retry_delay*4. And so on.
            This always includes jitter, where the final retry delay is a random number between 0 and the calculated retry delay.
            If retry_backoff is set to a number, that number is the maximum retry delay, in seconds.
        scheduled: epoch seconds for when the job should be scheduled, defaults to 0 (schedule right away)
        progress: job progress 0.0..1.0
        meta: arbitrary metadata to attach to the job
    Framework Set Properties
        attempts: number of attempts a job has had
        completed: job completion time epoch seconds
        queued: job enqueued time epoch seconds
        started: job started time epoch seconds
        touched: job touched/updated time epoch seconds
        results: payload containing the results, this is the return of the function provided, must be serializable, defaults to json
        error: stack trace if an runtime error occurs
        status: Status Enum, defaulst to Status.New
    """

    function: str
    kwargs: typing.Optional[dict] = None
    queue: typing.Optional[typing.Any] = None
    
    # queue: typing.Optional[typing.Union['TaskQueue', typing.Any]] = None
    key: typing.Optional[str] = Field(default_factory = get_default_job_key)
    timeout: typing.Optional[int] = Field(default_factory = settings.get_default_job_timeout)
    retries: typing.Optional[int] = Field(default_factory = settings.get_default_job_retries)
    ttl: typing.Optional[int] = Field(default_factory = settings.get_default_job_ttl)
    retry_delay: typing.Optional[float] = Field(default_factory = settings.get_default_job_retry_delay)

    retry_backoff: typing.Union[bool, float] = True
    
    heartbeat: int = 0
    scheduled: int = 0
    progress: float = 0.0
    attempts: int = 0
    completed: int = 0
    queued: int = 0
    started: int = 0
    touched: int = 0
    result: typing.Any = None
    error: typing.Optional[str] = None
    status: JobStatus = JobStatus.NEW
    meta: typing.Dict = Field(default_factory=dict)
    worker_id: typing.Optional[str] = None
    worker_name: typing.Optional[str] = None

    job_progress: typing.Optional[JobProgress] = Field(default_factory = JobProgress)

    # Allow for passing in a callback function to be called after the job is run
    job_callback: typing.Optional[typing.Union[str, typing.Callable]] = None
    job_callback_kwargs: typing.Optional[dict] = Field(default_factory=dict)

    if typing.TYPE_CHECKING:
        queue: typing.Optional[typing.Union['TaskQueue', typing.Any]] = None

    
    @validator("job_callback")
    def validate_job_callback(cls, v: typing.Optional[typing.Union[str, typing.Callable]]) -> typing.Optional[str]:
        """
        Validates the callback and returns the function name
        """
        return v if v is None else get_func_full_name(v)
    
    @property
    def job_callback_function(self) -> typing.Optional[typing.Callable]:
        """
        Returns the job callback function
        """
        if self.job_callback is None: return None
        func = import_function(self.job_callback)
        return ensure_coroutine_function(func)


    def __repr__(self):
        kwargs = ", ".join(
            f"{k}={v}"
            for k, v in {
                "id": self.id,
                "function": self.function,
                "has_callback": self.has_job_callback,
                "kwargs": self.kwargs,
                "queue": self.queue.queue_name,
                "scheduled": self.scheduled,
                "progress": self.progress,
                "process_ms": self.duration("process"),
                "start_ms": self.duration("start"),
                "total_ms": self.duration("total"),
                "attempts": self.attempts,
                "result": self.result,
                "error": self.error,
                "status": self.status,
                "meta": self.meta,
                "worker_id": self.worker_id,
                "worker_name": self.worker_name,
            }.items()
            if v is not None
        )
        return f"Job<{kwargs}>"
    
    @property
    def has_job_callback(self) -> bool:
        """
        Checks if the job has a callback
        """
        return self.job_callback is not None

    @property
    def short_repr(self):
        """
        Shortened representation of the job.
        """
        kwargs = ", ".join(
            f"{k}={v}"
            for k, v in {
                "id": self.id,
                "function": self.function,
                "kwargs": self.kwargs,
                "status": self.status,
                "attempts": self.attempts,
                "queue": self.queue.queue_name,
                "worker_id": self.worker_id,
                "worker_name": self.worker_name,
            }.items()
            if v is not None
        )
        return f"Job<{kwargs}>"
    

    @property
    def log_repr(self):
        """
        Shortened representation of the job.
        """
        kwargs = ", ".join(
            f"{k}={v}"
            for k, v in {
                "status": self.status,
                "attempts": self.attempts,
                "progress": self.progress,
                "kwargs": self.kwargs,
                "scheduled": self.scheduled,
                "process_ms": self.duration("process"),
                "start_ms": self.duration("start"),
                "total_ms": self.duration("total"),
                "error": self.error,
                "meta": self.meta,
            }.items()
            if v is not None
        )
        return f"{kwargs}"


    def __hash__(self):
        return hash(self.key)

    @property
    def id(self):
        """
        Returns the job id.
        """
        return self.queue.job_id(self.key)

    @classmethod
    def key_from_id(cls, job_id: str):
        """
        Returns the key from a job id.
        """
        return job_id.split(":")[-1]

    @property
    def abort_id(self):
        """
        Returns the abort id.
        """
        return f"{self.queue.abort_id_prefix}:{self.key}"

    def to_dict(self):
        """
        Serializes the job to a dictionary.
        """
        result = {}
        data = self.dict(
            exclude_none = True,
            exclude_defaults = True,
            exclude = {"kwargs"} if self.queue.serializer is not None else None,
        )
        for key, value in data.items():
            if key == "meta" and not value:
                continue
            if key == "queue" and value:
                value = value.queue_name
            result[key] = value
        if self.queue.serializer is not None: result["kwargs"] = self.kwargs
        return result


    def duration(self, kind) -> int:
        """
        Returns the duration of the job given kind.

        Kind can be process (how long it took to process),
        start (how long it took to start), or total.
        """
        if kind == "process":
            return self._duration(self.completed, self.started)
        if kind == "start":
            return self._duration(self.started, self.queued)
        if kind == "total":
            return self._duration(self.completed, self.queued)
        if kind == "running":
            return self._duration(now(), self.started)
        if kind == "queued":
            return self._duration(now(), self.queued)
        raise ValueError(f"Unknown duration type: {kind}")
    
    @property
    def job_duration(self) -> int:
        """
        Returns the duration of the job in ms.
        """
        for kind in {
            'process', 'total', 'start', 'running', 'queued'
        }:
            if duration := self.duration(kind):
                return duration
        
        # durations = []
        # for kind in {
        #     'running', 'start', 'queued'
        # }:
        #     if duration := self.duration(kind):
        #         durations.append(duration)
        #         # return duration
        # return max(durations, default=0)
        # return 0


    def _duration(self, a, b):
        #return a - b
        return a - b if a and b else None

    @property
    def stuck(self):
        """
        Checks if an active job is passed it's timeout or heartbeat.
        - if timeout is None, set timeout to 2 hrs = 7200.00
        - revised timeout to 30 mins = 1800.00
        """
        current = now()
        return (self.status == JobStatus.ACTIVE) and (
            seconds(current - self.started) > \
                # (self.timeout if self.timeout is not None else 1800.0)
                (self.timeout if self.timeout is not None else 7200.0)
            or (
                self.heartbeat and \
                    seconds(current - self.touched) > self.heartbeat
                )
        )

    def next_retry_delay(self):
        """
        Gets the next retry delay for the job.
        """
        if self.retry_backoff:
            max_delay = self.retry_delay
            if max_delay is True: max_delay = None
            return exponential_backoff(
                attempts = self.attempts,
                base_delay = self.retry_delay,
                max_delay = max_delay,
                jitter = True,
            )
        return self.retry_delay

    async def enqueue(self, queue: 'TaskQueue' = None):
        """
        Enqueues the job to it's queue or a provided one.

        A job that already has a queue cannot be re-enqueued. Job uniqueness is determined by its id.
        If a job has already been queued, it will update it's properties to match what is stored in the db.
        """
        queue = queue or self.queue
        assert queue, "Queue unspecified"
        if not await queue.enqueue(self):
            await self.refresh()

    async def abort(self, error: typing.Any, ttl: int = 5):
        """Tries to abort the job."""
        await self.queue.abort(self, error = error, ttl = ttl)

    async def finish(self, status: JobStatus, *, result: typing.Any = None, error: typing.Any = None):
        """Finishes the job with a Job.Status, result, and or error."""
        await self.queue.finish(self, status = status, result = result, error = error)

    async def retry(self, error: typing.Any):
        """Retries the job by removing it from active and requeueing it."""
        await self.queue.retry(self, error)

    async def update(self, **kwargs):
        """
        Updates the stored job in keydb.

        Set properties with passed in kwargs.
        """
        for k, v in kwargs.items():
            setattr(self, k, v)
        await self.queue.update(self)
    
    # async def reset(self):
    #     """
    #     Resets the job to it's initial state and rerun
    #     """
    #     await self.queue.reset(self)
    
    async def set_progress(self, total: typing.Optional[int] = None, completed: typing.Optional[int] = None):
        """
        Sets the job's progress.
        """
        if total is not None:
            self.job_progress.total = total
        if completed is not None:
            self.job_progress.completed = completed
        await self.queue.update(self)
    
    async def incr_progress(self, incr: typing.Optional[int] = 1):
        """
        Increments the job's progress.
        """
        self.job_progress.completed += incr
        await self.queue.update(self)

    @property
    def _progress(self) -> typing.Optional[float]:
        """
        Returns the job's progress as a float.
        """
        return self.job_progress.progress

    async def refresh(self, until_complete: int = None):
        """
        Refresh the current job with the latest data from the db.

        until_complete: None or Numeric seconds. if None (default), don't wait,
            else wait seconds until the job is complete or the interval has been reached. 0 means wait forever
        """
        job = await self.queue.job(self.key)

        if not job: raise RuntimeError(f"{self} doesn't exist")
        self.replace(job)
        if until_complete is not None and not self.completed:
            async def callback(_id, status):
                if status in TERMINAL_STATUSES:
                    return True

            await self.queue.listen([self.key], callback, until_complete)
            await self.refresh()

    @lazyproperty
    def job_fields(self) -> typing.List[str]:
        """
        Returns the fields of the job.
        """
        return [field.name for field in self.__fields__.values()]

    def replace(self, job: 'Job'):
        """
        Replace current attributes with job attributes.
        """
        for field in job.job_fields:
            setattr(self, field, getattr(job, field))

    
    @classmethod
    def get_fields(cls):
        """
        Returns the fields of the job.
        """
        return [field.name for field in cls.__fields__.values()]
    
    @classmethod
    def from_kwargs(cls, job_or_func: typing.Union[str, 'Job', typing.Callable], **kwargs) -> 'Job':
        """
        Returns a job from kwargs.
        """
        job_kwargs = {"kwargs": {}}
        job_fields = cls.get_fields()
        for k, v in kwargs.items():
            # Allow passing underscored keys
            # to prevent collision with job fields
            if k.startswith('_') and k[1:] in job_fields:
                job_kwargs[k[1:]] = v
            elif k in job_fields:
                job_kwargs[k] = v
            else:
                job_kwargs["kwargs"][k] = v
        
        if isinstance(job_or_func, type(cls)):
            job = job_or_func
            for k, v in job_kwargs.items():
                setattr(job, k, v)

        elif isinstance(job_or_func, str) or callable(job_or_func):
            job = cls(function = get_func_name(job_or_func), **job_kwargs)

        else:
            raise ValueError(f"Invalid job type {type(job_or_func)}")
    
        return job
        
    async def _run_job_callback(self):
        """
        Runs the job callback
        """
        if not self.has_job_callback: return
        if self.status not in TERMINAL_STATUSES: return
        try:
            await self.job_callback_function(
                status = self.status,
                job_id = self.id,
                function_name = self.function,
                duration = self.job_duration,
                result = self.result,
                error = self.error,
                **self.job_callback_kwargs or {},
            )
        except Exception as e:
            logger.error(f"Failed to run job callback for {self}: {e}")

    async def run_job_callback(self):
        """
        Runs the job callback in the background
        """
        create_background_task(self._run_job_callback())

    """
    Queue Keys for Job
    """

    @lazyproperty
    def queued_key(self) -> str:
        if self.worker_id:
            return f"{self.queue.queued_key}:{self.worker_id}"
        if self.worker_name:
            return f"{self.queue.queued_key}:{self.worker_name}"
        return self.queue.queued_key
    
    @lazyproperty
    def active_key(self) -> str:
        if self.worker_id:
            return f"{self.queue.active_key}:{self.worker_id}"
        if self.worker_name:
            return f"{self.queue.active_key}:{self.worker_name}"
        return self.queue.active_key
    
    @lazyproperty
    def incomplete_key(self) -> str:
        if self.worker_id:
            return f"{self.queue.incomplete_key}:{self.worker_id}"
        if self.worker_name:
            return f"{self.queue.incomplete_key}:{self.worker_name}"
        return self.queue.incomplete_key

# Job.update_forward_refs()